#Title:          BuildConsensusSpectrum.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
BuildConsensusSpectrum:
- Given many similar spectra which we consider to be the "same" (either due to
Inspect annotations or due to clustering), we'd like to generate a CONSENSUS
SPECTRUM.  The consensus spectrum should contain less noise than the individual
spectra, and mass errors should be decreased.  Also, the computation shouldn't
require too much time; for efficiency, it should require just one I/O pass through
the spectra.

Current pseudocode:

# Accumulate intensity and peak count:
For spectrum S:
  Read in peaks from disk
  Intensity1 = max(max peak intensity, grass peak intensity * 20)
  For peak P in spectrum S:
    ScaledIntensity = (Intensity(P) / Intensity1)
    MassBin = mass of P, rounded to nearest 0.1Da
    TotalIntensity[MassBin] += ScaledIntensity
    PeakCount[MassBin]++
    PeakCount[MassBin-1]++
    PeakCount[MassBin+1]++
    
# Process intensity into a peak list:
Iterate over peaks P from high to low peak-count:
  if Assimilated[P], continue
  Peak P assimilates neighboring peaks if their total intensity is lower
  new spectrum receives this assimilated peak
"""
import PyInspect
import MSSpectrum
import os
import sys
import traceback
import cPickle

USE_COUNT_FLAG = 1

# Scaling factors, for levels of peak presence from 0% to 100%.
ScalingFactors = {}
for X in range(0, 101):
    #ScalingFactors[X] = 0.95 + 0.05 * (1.0 + X / 100.0)**5
    ScalingFactors[X] = 0.95 + 0.05 * (1.0 + X / 100.0)**5
    #print X, ScalingFactors[X]


class ConsensusBuilder:
    def __init__(self, Charge = None):
        self.SpectrumCount = 0
        self.TotalMZ = 0
        self.Intensity = {}
        self.PeakCount = {}
        self.Charge = Charge
        self.SignalPeakCount = 0
    def DebugPrint(self):
        Bins = self.Intensity.keys()
        Bins.sort()
        MinBin = min(Bins)
        MaxBin = max(Bins)
        for Bin in range(MinBin, MaxBin + 1):
            Str = "%s\t%s\t%s\t"%(Bin, self.Intensity.get(Bin, 0), self.PeakCount.get(Bin, 0))
            print Str
        
    def AddSpectrum(self, Spectrum):
        """
        Add one more spectrum to the binned intensity and peak counts
        """
        self.TotalMZ += Spectrum.PrecursorMZ
        self.SpectrumCount += 1
        if not self.Charge:
            self.Charge = Spectrum.Charge # ASSUMED: all spectra have same charge!
        if not self.SignalPeakCount:
            # Expect to see this many "signal" peaks:
            self.SignalPeakCount = int(round((Spectrum.ParentMass / 100.0) * 4))
        MaxIntensity = 0
        IntensityList = []
        for Peak in Spectrum.Peaks:
            MaxIntensity = max(MaxIntensity, Peak.Intensity)
            IntensityList.append(Peak.Intensity)
        IntensityList.sort()
        # Spectra with high signal-to-noise are weighted a bit more heavily:
        if len(IntensityList) > self.SignalPeakCount:
            MinimumPresenceIntensity = IntensityList[-self.SignalPeakCount]
            ScalingFactor = min(20.0 / MaxIntensity, 1.0 / MinimumPresenceIntensity)
        else:
            ScalingFactor = 20.0 / MaxIntensity
            MinimumPresenceIntensity = 0
        #print "%s peaks; Spectrum has scaling factor %s (grass %s, max peak %s)"%(len(IntensityList), ScalingFactor, MinimumPresenceIntensity, MaxIntensity)
        for Peak in Spectrum.Peaks:
            MZBin = int(round(Peak.Mass * 10))
            ScaledIntensity = Peak.Intensity * ScalingFactor
            self.Intensity[MZBin] = self.Intensity.get(MZBin, 0) + ScaledIntensity
            if Peak.Intensity > MinimumPresenceIntensity:
                self.PeakCount[MZBin] = self.PeakCount.get(MZBin, 0) + 1
                self.PeakCount[MZBin-1] = self.PeakCount.get(MZBin-1, 0) + 1
                self.PeakCount[MZBin-2] = self.PeakCount.get(MZBin-2, 0) + 1
                self.PeakCount[MZBin+1] = self.PeakCount.get(MZBin+1, 0) + 1
                self.PeakCount[MZBin+2] = self.PeakCount.get(MZBin+2, 0) + 1
            #self.Intensity[MZBin] = self.Intensity.get(MZBin, 0) + ScaledIntensity
    def PickleCluster(self, PicklePath):
        """
        Serialize the information we've read from many spectra.  This method is used
        if we want to reserve the option to add to the cluster later.
        """
        File = open(PicklePath, "wb")
        cPickle.dump(self.Charge, File)
        cPickle.dump(self.TotalMZ, File)
        cPickle.dump(self.SpectrumCount, File)
        cPickle.dump(self.Intensity, File)
        cPickle.dump(self.PeakCount, File)
        File.close()
    def UnpickleCluster(self, PicklePath):
        """
        Sister method to PickleCluster - load a cluster from disk.
        """
        File = open(PicklePath, "rb")
        self.Charge = cPickle.load(File)
        self.TotalMZ = cPickle.load(File)
        self.SpectrumCount = cPickle.load(File)
        self.Intensity = cPickle.load(File)
        self.PeakCount = cPickle.load(File)
        File.close()
        
    def ProduceConsensusSpectrum(self):
        Spectrum = MSSpectrum.SpectrumClass()
        Spectrum.Charge = self.Charge
        Spectrum.PrecursorMZ = self.TotalMZ / float(max(self.SpectrumCount, 1))
        Spectrum.ParentMass = (Spectrum.PrecursorMZ * Spectrum.Charge) - (Spectrum.Charge - 1)*1.0078
        # Iterate over intensity entries:
        self.AssimilationFlag = {}
        SortedList = []
        for (Bin, Score) in self.Intensity.items():
            SortedList.append((Score, Bin))
        SortedList.sort()
        SortedList.reverse()
        Spectrum.Peaks = []
        for (Intensity, Bin) in SortedList:
            if self.AssimilationFlag.get(Bin, 0):
                continue
            Peak = MSSpectrum.PeakClass(Bin / 10.0, Intensity)
            for NearBin in (Bin-1, Bin-2, Bin-3, Bin+1, Bin+2, Bin+3):
                if self.AssimilationFlag.get(NearBin, 0):
                    continue
                self.AssimilationFlag[NearBin] = 1
                Peak.Intensity += self.Intensity.get(NearBin, 0)
            # Scale the intensity by the peak count, IF the cluster is large:
            if USE_COUNT_FLAG and self.SpectrumCount > 4:
                FractionPresent = self.PeakCount.get(Bin,0) / float(self.SpectrumCount)
                Bin = min(100, int(round(FractionPresent * 100)))
                #print FractionPresent, Bin, ScalingFactors[Bin]
                Peak.Intensity *= ScalingFactors[Bin]
            Spectrum.Peaks.append(Peak)
        Spectrum.Peaks.sort()
        return Spectrum
    def AssimilateCluster(self, OtherCluster):
        """
        Assimilate the other consensus information into our consensus spectrum.
        """
        self.SpectrumCount += OtherCluster.SpectrumCount
        self.TotalMZ += OtherCluster.TotalMZ
        for (Key, Value) in OtherCluster.Intensity.items():
            self.Intensity[Key] = self.Intensity.get(Key, 0) + Value
        for (Key, Value) in OtherCluster.PeakCount.items():
            self.PeakCount[Key] = self.PeakCount.get(Key, 0) + Value
def TestConsensus(AnnotationsFile, Annotation, Charge = 2):
    """
    Build a consensus spectrum for a collection of spectra, and verify
    that it looks good.
    """
    #AnnotationsFile = "ConsensusTest.txt"
    #Annotation = "M.D+173VTIQHPWFK.R"
    SpectrumDir = "e:\\ms\\lens\\spectra"
    TestOutputFile = "ConsensusTest.dta"
    Builder = ConsensusBuilder()
    InputFile = open(AnnotationsFile, "rb")
    BestScores = []
    #MGFFile = open("TestTest.mgf", "wb") #%TEMP
    #MGFFile.close() #%TEMP
    for FileLine in InputFile.xreadlines():
        Bits = FileLine.split("\t")
        try:
            FilePath = Bits[0]
            FilePos = int(Bits[15])
            SpectrumCharge = int(Bits[4])
            MQScore = float(Bits[5])
        except:
            continue
        if SpectrumCharge != Charge:
            continue
        FilePath = os.path.join(SpectrumDir, FilePath.replace("/", "\\").split("\\")[-1])
        BestScores.append((MQScore, FilePath, FilePos))
        Spectrum = MSSpectrum.SpectrumClass()
        Spectrum.ReadPeaks("%s:%s"%(FilePath, FilePos))
        Spectrum.SetCharge(Charge)
        Builder.AddSpectrum(Spectrum)
        # Try ari's thingy:
        #Spectrum.WriteMGFPeaks("TestTest.mgf") #%TEMP
    InputFile.close()
    print "Build consensus spectrum for %s members."%Builder.SpectrumCount
    Consensus = Builder.ProduceConsensusSpectrum()
    Consensus.WritePeaks(TestOutputFile)
    #Command = "MakeConsensus -mgf TestTest.mgf > %s"%(TestOutputFile) #%TEMP
    #print Command #%TEMP
    #os.system(Command) #%TEMP
    PySpectrum = PyInspect.Spectrum(TestOutputFile, 0)
    Results = PySpectrum.ScorePeptideDetailed(Annotation)
    ConsensusScore = Results[0]
    print "Consensus spectrum score: %.2f (%.2f, %.2f, %.2f, %d)"%(ConsensusScore, Results[1], Results[2], Results[3], Results[4])
    # Compare the CONSENSUS score to the average for the BEST FIVE:
    BestScores.sort()
    TopHits = 0
    for (Score, Path, Pos) in BestScores[-5:]:
        PySpectrum = PyInspect.Spectrum(Path, Pos)
        Score = PySpectrum.ScorePeptide(Annotation)
        TopHits += Score
    BestFiveAverage = TopHits / 5.0
    ScoreGain = ConsensusScore - BestFiveAverage
    print "Consensus %s vs top-5 average %s (%s)"%(ConsensusScore, BestFiveAverage, ScoreGain)
    return ScoreGain

def TestMain():
    TestCases = [("Consensus.GNTIEIQGDDAPSLWVYGFSDR.txt", "K.GNTIEIQGDDAPSLWVYGFSDR.V", 2),
                 ("ConsensusTest.txt", "-.M+42DVTIQHPWFK.R", 1),
                 ("ConsensusTest.txt", "-.M+42DVTIQHPWFK.R", 2),
                 ("Consensus.R.QD-17DHGYISR.E.txt", "R.Q-17DDHGYISR.E", 1),
                 ("Consensus.R.QD-17DHGYISR.E.txt", "R.Q-17DDHGYISR.E", 2),
                 ]
    ResultTotal = 0
    ResultCount = 0
    #TestCases = TestCases[0:1] # TEMP%!
    for (AnnotationFile, Annotation, Charge) in TestCases:
        ResultTotal += TestConsensus(AnnotationFile, Annotation, Charge)
        ResultCount += 1
    print "OVERALL RESULTS: Average MQGain is %s"%(ResultTotal / float(max(1, ResultCount)))
    
if __name__ == "__main__":
    # Given the filename of a cluster, print verbose info:
    FileName = sys.argv[1]
    Bob = ConsensusBuilder()
    Bob.UnpickleCluster(FileName)
    Bob.DebugPrint()



#Title:          PTMAnalysis.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""
PTMAnalysis.py
This wrapper script automates the steps involved in processing raw Insepct
results into finalized PTM site identifications. 
1. FDRUtils.py - obtain a consistent pvalue across all runs
2. ComputePTMFeatures.py - group individual annotations into peptides and
    compute some features for each.
3. BuildMGF.py - builds a single .mgf file out of all the consensus spectra
    created in step 2 (in preparation for the Inspect run)
4. Inspect run.  Search clustered spectra (unmodified) against a large database
5. PTMSearchBigDB.py - Integrates the results of the inspect search against a
    large database, findig a delta-score.
6. TrainPTMFeatures.py - Computes the PValue of each site from a model
7. AdjustPTM.py - attempts to merge and reconcile sites
8. AdjustPTM.py (known chemistry) - attempts to find known explanations for
    the site

Depending on the size of your dataset, this program may take quite a while
(up to 1 day per million spectra).  It is reasonably easy to parallelize
the time-intensive steps (2) and (4) above; doing so is up to the user, since
compute clusters are so heterogenous.

To perform only a subset of the steps, you can designate start and stop steps.
This is a useful feature if perhaps the program crashes at some point, or if
you would like to run some steps on a grid, and some locally.
To start from a step other than one, the program assumes that all the previous
steps have been executed, and that their outputs are in the proper directories.
See the functions below for the expected directory and file names.
"""
import os
import getopt
import sys
import traceback
from Utils import *

UsageInfo = """
PTMAnalysis.py - produces a set of PTM sites.

Required options:
 -r [FileName] - The name of the results file to parse.  If a directory is
    specified, then all .txt files within the directory will be combined into
    one report
 -d [FileName] - The name of the database file (.trie format) searched.
 -w [FileName] - The final output filename
 -s [Dir] - Directory containing the MS/MS spectra
 -B [FilePath] - Path to the large database (for unmodified "decoy" search).
 -S [Value] - The fraction of sequences in the database that are shuffled.
 -k [FileName]: Known chemistry filename.  If specified, consider altering
   sites to match known chemical adducts; report the best site-score
   attainable by using known chemical adducts.

Additional options:
 -p [Value] - The pvalue cutoff for spectra annotations.  Default to 0.1
 -q [Value] - The pvalue cutoff for PTM site annotations. Default to 0.05
 -t [Path] - The path where all intermediate results will be written.
        Default to PTMTempFiles
 -i [Instrument] - The type of instrument that the Spectra were created on.
     Default is ESI-ION-TRAP.  Valid values are: QTOF, FT-Hybrid
 -n [FileName] The parameters file to use in the BigSearch (spectra and DB will be replaced, blind will be removed)
     
Advanced options: to run only a subset of the steps
 -x [StartStep] - Start with step X (assume that previous steps are done)
 -y [StopStep] - Stop with step Y (inclusive)

Protein selection can be performed, replacing the protein identification
with a parsimonious set of protein IDs (using a simple iterative
approach).  The following options are required for protein selection:
 -a: Replace protein identifications with a "parsimonious" set of protein IDs.

"""

class Steps:
    First = 1
    PValue = 1
    ComputePTMFeatures = 2
    BuildMGF = 3
    RunInspect = 4
    SearchBigDB = 5
    TrainPTMFeatures = 6
    AdjustPeptides = 7
    AdjustToKnownChemistry = 8
    Last = 8

class WrapperClass:
    def __init__ (self):
        self.SpectrumPValue = 0.1 # for the pvalue.py program
        self.SitePValue = 0.05 # for the final output
        self.BasePath = "PTMTempFiles"
        self.PValuePath = None
        self.PValueOutput = None
        self.ComputePTMPath = None
        self.ComputePTMOutput = None
        self.TrainPTMPath = None
        self.TrainPTMOutput = None
        self.TrainPTMModelOutput = None
        self.AdjustDir = None
        self.AdjustOutputPath = None
        self.AdjustModelOutput = None
        self.SearchBigDBPath = None
        self.SearchBigDBOutput = None
        self.InputResults = None
        self.DatabaseFile = None
        self.PercentShuffled = None
        self.SpectraDir = None
        self.SelectProteins = 0 #default
        self.FinalOutputFile = None
        self.BigDB = None
        self.Instrument = "ESI-ION-TRAP"
        self.BuildMGFPath = None
        self.InspectOutDir = None
        self.MGFPath = None
        self.SpawnFlag = 0
        self.StartStep = Steps.First
        self.StopStep = Steps.Last
        self.KnownChemistryFileName = None
        self.TrainPTMModelType = "svm" # default
        self.ParamsFile = None
    def SetupDirectories(self):
        """
        Below the basepath there will be a group of directories, one for
        each major step
        """
        self.PValuePath = os.path.join(self.BasePath, "PValued")
        self.ComputePTMPath = os.path.join(self.BasePath, "ComputePTMFeatures")
        self.TrainPTMPath = os.path.join(self.BasePath, "TrainPTMFeatures")
        self.AdjustDir = os.path.join(self.BasePath, "AdjustPTM")
        self.SearchBigDBPath = os.path.join(self.BasePath, "SearchBigDB")
        self.BuildMGFPath = os.path.join(self.BasePath, "BuildMGF")
        self.InspectOutDir = os.path.join(self.BasePath, "InspectOut")
        print "Making temporary directories in %s for all intermediate output"%self.BasePath
        MakeDirectory(self.PValuePath)
        MakeDirectory(self.ComputePTMPath)
        MakeDirectory(self.TrainPTMPath)
        MakeDirectory(self.AdjustDir)
        MakeDirectory(self.SearchBigDBPath)
        MakeDirectory(self.BuildMGFPath)
        MakeDirectory(self.InspectOutDir)
    def RunPValue(self):
        """
        FDRUtils.py
            -r InputResults
            -w OutputResults
            -S Percent of database shuffled (optional)
            -p pvalue cutoff
            -s Distribution file
            -i Distribution image
            -H write out results from shuffled protein
        """
        self.PValueOutput = self.PValuePath # default, a directory for directory input
        DistributionFile = os.path.join(self.PValuePath, "Distribution.txt")
        if not os.path.isdir(self.InputResults): # the InputResults is a single file
            FileName = os.path.split(self.InputResults)[1]
            self.PValueOutput = os.path.join(self.PValueOutput, FileName)
        PValueArgs = ""
        if self.PercentShuffled:
            PValueArgs = "-r %s -w %s -S %f -p %f -s %s -i -H" %(self.InputResults, self.PValueOutput, self.PercentShuffled, self.SpectrumPValue, DistributionFile)
        else:
            PValueArgs = "-r %s -w %s -p %f -s %s -i -H" %(self.InputResults, self.PValueOutput, self.SpectrumPValue, DistributionFile)
        if self.SelectProteins:
            PValueArgs += " -a -d %s"%self.DatabaseFile
        PValueArgs += " -b "
        if self.StartStep <= Steps.PValue and Steps.PValue <= self.StopStep:
            print "Step %s: FDRUtils"%Steps.PValue
            print "Arguments: %s"%PValueArgs
            if self.SpawnFlag:
                Command = "python FDRUtils.py %s"%PValueArgs
                print Command
                os.system(Command)
            else:
                import FDRUtils
                ArgsList = PValueArgs.split()
                Parser = FDRUtils.PValueParser()
                Parser.ParseCommandLine(ArgsList)
                FDRUtils.Main(Parser)
                del FDRUtils
        else:
            print "Skipping Step %s: FDRUtils"%Steps.PValue
    def RunComputePTMFeatures(self):
        """
        ComputePTMFeatures.py
            -r InputResults
            -w OutputDir
            -d Database
            -s spectra
        """
        self.ComputePTMOutput = os.path.join(self.ComputePTMPath, "PTMFeatures.txt")
        Args = " -r %s -w %s -d %s -s %s"%(self.PValueOutput, self.ComputePTMPath, self.DatabaseFile, self.SpectraDir)
        if self.StartStep <= Steps.ComputePTMFeatures and Steps.ComputePTMFeatures <= self.StopStep:
            print "Step %s: ComputePTMFeatures"%Steps.ComputePTMFeatures
            print "Arguments: %s"%Args
            if self.SpawnFlag:
                Command = "ComputePTMFeatures.py %s"%Args
                print Command
                os.system(Command)
            else:
                import ComputePTMFeatures
                ArgsList = Args.split()
                Computer = ComputePTMFeatures.PTMFeatureComputer()
                Computer.ParseCommandLine(ArgsList)
                Computer.ComputeFeaturesMain()
                del ComputePTMFeatures
        else:
            print "Skipping Step %s: ComputePTMFeatures"%Steps.ComputePTMFeatures
    def RunBuildMGF(self):
        """
        BuildMGF.py
            -d PTM feature directory
            -m MGF file to make
        """
        self.MGFPath = os.path.join(self.BuildMGFPath, "spectra.mgf")
        Args = " -d %s -m %s"%(self.ComputePTMPath, self.MGFPath)
        
        if self.StartStep <= Steps.BuildMGF and Steps.BuildMGF <= self.StopStep:
            print "Step %s: BuildMGF"%Steps.BuildMGF
            print "Arguments:  %s"%Args
            if self.SpawnFlag:
                Command = "BuildMGF.py %s"%Args
                print Command
                os.system(Command)
            else:
                ArgsList = Args.split()
                import BuildMGF
                Builder = BuildMGF.MGFBuilder()
                Builder.ParseCommandLine(ArgsList)
                Builder.Main()
                del BuildMGF
        else:
            print "Skipping Step %s: BuildMGF"%Steps.BuildMGF
    def RunInspect(self):
        """
        Given that the mgf file was previously created, here we create an
        input file for Inspect and run it.
        """
        InspectExe = None
        if sys.platform == "win32":
            InspectExe = "Inspect.exe"
        else:
            InspectExe = "./inspect"
        InspectIn = "BigSearch.in"
        self.InspectOutFile = os.path.join(self.InspectOutDir, "Results.txt")
        Command = "%s -i %s -o %s"%(InspectExe, InspectIn, self.InspectOutFile)
        
        Dict = {}
        if self.ParamsFile:
            File = open(self.ParamsFile,'r')
            
            for Line in File:
                Line = Line.strip()
                if Line == "":
                    continue
                Bits = Line.split(",")
                if Dict.has_key(Bits[0].lower()):
                    Dict[Bits[0].lower()].append((",".join(Bits[1:])).lower())
                else:
                    Dict[Bits[0].lower()] = [(",".join(Bits[1:])).lower()]
            File.close()
            if Dict.has_key("blind"):
                del Dict["blind"]
            if Dict.has_key("mods"):
                del Dict["mods"]
            if Dict.has_key("sequencefile"):
                del Dict["sequencefile"]
            if Dict.has_key("unrestrictive"):
                del Dict["unrestrictive"]
            if Dict.has_key("maxptmsize"):
                del Dict["maxptmsize"]
            
        else:
            
            Dict["protease"] = ["Trypsin"]
            Dict["mod"] = ["57,C,fix"]
            Dict["tagcount"] = ["25"]

        if not Dict.has_key("instrument"):
            Dict["instrument"] = [self.Instrument]
        Dict["db"] = [self.BigDB]
        Dict["spectra"] = [self.MGFPath]
        
        InFileCommands = ""
        for Key in Dict.keys():
            List = Dict[Key]
            Str = ""
            for L in List:
                Str += "%s,%s\n"%(Key,L)
            InFileCommands += Str

        #InFileCommands = "spectra,%s\n"%self.MGFPath
        #InFileCommands += "instrument,%s\n"%self.Instrument
        #InFileCommands += "protease,Trypsin\n"
        #InFileCommands += "DB,%s\n"%self.BigDB
        #InFileCommands += "mod,57,C,fix\n"
        #InFileCommands += "tagcount,25\n"
        #print InFileCommands
        #raw_input()
        Handle = open(InspectIn, "wb")
        Handle.write(InFileCommands)
        Handle.close()
        if self.StartStep <= Steps.RunInspect and Steps.RunInspect <= self.StopStep:
            print "Step %s: Run Inspect, searching consensus spectra"%Steps.RunInspect
            print "Arguments: %s"%Command
            os.system(Command)
        else:
            print "Skipping Step %s: Run Inspect, searching consensus spectra"%Steps.RunInspect
    def RunPTMSearchBigDB(self):
        """
        PTMSearchBigDB.py
            -d PTM feature directory
            -w Outputfile to write
            -r Inspect Search Results (default to directory, not file)
        """
        self.SearchBigDBOutput = os.path.join(self.SearchBigDBPath, "Results.txt")
        # we use the InspectOutDir in case there are multiple files within the directory.
        Args = " -d %s -w %s -r %s"%(self.ComputePTMPath, self.SearchBigDBOutput, self.InspectOutDir)
        if self.StartStep <= Steps.SearchBigDB and Steps.SearchBigDB <= self.StopStep:
            print "Step %s: PTMSearchBigDB"%Steps.SearchBigDB
            print "Arguments:  %s"%Args
            if self.SpawnFlag:
                Command = "PTMSearchBigDB.py %s"%Args
                print Command
                os.system(Command)
            else:
                ArgsList = Args.split()
                import PTMSearchBigDB
                Searcher = PTMSearchBigDB.PTMSearcher()
                Searcher.ParseCommandLine(ArgsList)
                Searcher.Main()
                del PTMSearchBigDB
        else:
            print "Skipping Step %s: PTMSearchBigDB"%Steps.SearchBigDB
    def RunTrainPTMFeatures(self):
        """
        TrainPTMFeatures.py
            -u InputResults
            -v OutputResults
            -m ModelType
            -w OutputModel
        """
        self.TrainPTMOutput = os.path.join(self.TrainPTMPath, "Results.txt")
        self.TrainPTMModelOutput = os.path.join(self.TrainPTMPath, "model.%s.txt"%self.TrainPTMModelType)
        Args = "-u %s -v %s -m %s -w %s"%(self.SearchBigDBOutput, self.TrainPTMOutput, self.TrainPTMModelType, self.TrainPTMModelOutput)
        if self.StartStep <= Steps.TrainPTMFeatures and Steps.TrainPTMFeatures <= self.StopStep:
            print "Step %s: TrainPTMFeatures"%Steps.TrainPTMFeatures
            print "Arguments: %s"%Args
            if self.SpawnFlag:
                Command = "TrainPTMFeatures.py %s"%Args
                print Command
                os.system(Command)
            else:
                ArgsList = Args.split()
                import TrainPTMFeatures
                Trainer = TrainPTMFeatures.PTMFeatureTrainer()
                Trainer.ParseCommandLine(ArgsList)
                Trainer.TrainModel()
                del TrainPTMFeatures
        else:
            print "Skipping Step %s: TrainPTMFeatures"%Steps.TrainPTMFeatures
    def AdjustPeptides(self):
        """
        AdjustPTM.py
            -r InputResults
            -w OutputResults
            -d Database File
            -c Cluster directory from ComputePTMFeatures
            -m model INPUT filename
            -M model OUTPUT filename
        """
        self.AdjustOutputPath = os.path.join(self.AdjustDir, "Results.txt")
        self.AdjustModelOutput = os.path.join(self.AdjustDir, "model.%s.txt"%self.TrainPTMModelType)
        Args = "-r %s -w %s -d %s -c %s -m %s -M %s -z "%(self.TrainPTMOutput, self.AdjustOutputPath, self.DatabaseFile, self.ComputePTMPath, self.TrainPTMModelOutput, self.AdjustModelOutput)
        if self.StartStep <= Steps.AdjustPeptides and Steps.AdjustPeptides <= self.StopStep:
            print "Step %s: AdjustPTM"%Steps.AdjustPeptides
            print "Arguments: %s"%Args
            if self.SpawnFlag:
                Command = "AdjustPTM.py %s"%Args
                print Command
                os.system(Command)
            else:
                ArgsList = Args.split()
                import AdjustPTM
                Adjutant = AdjustPTM.PTMAdjuster()
                Adjutant.ParseCommandLine(ArgsList)
                Adjutant.Main()
                del AdjustPTM
        else:
            print "Skipping Step %s: AdjustPTM"%Steps.AdjustPeptides
    def AdjustToKnownChemistry(self):
        """
        AdjustPTM.py
            -r InputResults
            -w OutputResults
            -m input Model
            -d Database
            -c Clusters 
            -k Known PTM file
            -v verbose output file
        """
        if not self.KnownChemistryFileName:
            print "* Skipping AdjustToKnownChemistry: Requires a file (-k) of 'common' modifications."
            return
        KnownPTMVerboseOutputPath = os.path.join(self.BasePath, "KnownPTMOutput.txt")
        Args = "-r %s -w %s -m %s -d %s -c %s -k %s -v %s"%(self.AdjustOutputPath, self.FinalOutputFile, self.AdjustModelOutput, self.DatabaseFile, self.ComputePTMPath, self.KnownChemistryFileName, KnownPTMVerboseOutputPath)
        
        if self.StartStep <= Steps.AdjustToKnownChemistry and Steps.AdjustToKnownChemistry <= self.StopStep:
            print "Step %s: AdjustToKnownChemistry"%Steps.AdjustToKnownChemistry
            print "Arguments: %s"%Args
            if self.SpawnFlag == 1:
                Command = "AdjustPTM.py %s"%Args
                print Command
                os.system(Command)
            else:
                ArgsList = Args.split()
                import AdjustPTM
                Adjutant = AdjustPTM.PTMAdjuster()
                Adjutant.ParseCommandLine(ArgsList)
                Adjutant.Main()
                del AdjustPTM
        else:
            print "Skipping Step %s: AdjustToKnownChemistry"%Steps.AdjustToKnownChemistry
    def ParseCommandLine(self, Arguments):
        "Args is a list of arguments only (does not include sys.argv[0] == script name)"
        (Options, Args) = getopt.getopt(Arguments, "r:d:w:s:p:q:t:S:B:i:x:y:k:bZm:n:")
        OptionsSeen = {}
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-r":
                self.InputResults = Value
            elif Option == "-m":
                self.TrainPTMModelType = Value # "svm" or "lda"
            elif Option == "-d":
                self.DatabaseFile = Value
            elif Option == "-w":
                self.FinalOutputFile = Value
            elif Option == "-s":
                self.SpectraDir = Value
            elif Option == "-p":
                self.SpectrumPValue = float(Value)
            elif Option == "-q":
                self.SitePValue = float(Value)
            elif Option == "-t":
                self.BasePath = Value
            elif Option == "-S":
                self.PercentShuffled = float (Value)
            elif Option == "-B":
                self.BigDB = Value
            elif Option == "-x":
                self.StartStep = int (Value)
            elif Option == "-y":
                self.StopStep = int (Value)
            elif Option == "-k":
                self.KnownChemistryFileName = Value
            elif Option == "-Z":
                self.SpawnFlag = 1
            elif Option == "-n":
                self.ParamsFile = Value
            else:
                print "** Unknown option:", Option, Value
        if not OptionsSeen.has_key("-r") or not OptionsSeen.has_key("-d") or not OptionsSeen.has_key("-w") or not OptionsSeen.has_key("-s") or not OptionsSeen.has_key("-S"):
            print "Missing required options (r, d, w, s, S)"
            print UsageInfo
            sys.exit(-1)
        if not self.BigDB:
            print "Missing large DB (-B)"
            print UsageInfo
            sys.exit(-1)

def Main():
    "Main control box for the script"
    Wrap = WrapperClass()
    Wrap.ParseCommandLine(sys.argv[1:])
    Wrap.SetupDirectories()
    #now we run the parts of the scripts one after another
    print "\n*** Starting to run components ***"
    Wrap.RunPValue()
    Wrap.RunComputePTMFeatures()
    Wrap.RunBuildMGF()
    Wrap.RunInspect()
    Wrap.RunPTMSearchBigDB()
    Wrap.RunTrainPTMFeatures()
    Wrap.AdjustPeptides()
    Wrap.AdjustToKnownChemistry()
    

if __name__ == "__main__":
    try:
        import psyco
        psyco.full()
    except:
        print "psyco not found - running without optimization"
    Main()    
        



#Title:          PLSUtils.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""PLSUtils.py

These are a set of related functions which help calculate the Phosphate
Localization Score or PLS.  The differences between this and the Ascore
are described in Albuquerque MCP 2008.
1. The winner and runner up are determined by the MQScore, not peptide score.
2. We do not repeatedly filter the peaks to find the optimal
AScore.  That just takes too long.  We use the default Inspect peak filtering which
leaves ~12 peaks / 100 m/z.  This corresponds to a p of 0.12 for the
binomial.
3. We will not do any thing for peptides which contain
more than 2 sites of phosphorylation.  I don't trust those annotations
anyway.
"""

import math
import copy
from Utils import*
Initialize()

class PLSClass:
    def __init__(self):
        self.ChooseTable = {} # (N, k) -> value of NchooseK
        self.Factorial = {}
        self.Factorial[0] = 1

    def ComputeBinomial(self, N, n, p=0.12):
        """ Make sure that you have populated the Choose table and Factorial table
            I have defaulted the p = 0.12 because it's inspect peak density.
        """
        #print "computing binomial with N %s and n %s"%(N,n)
        Sum = 0
        for k in range(n, N+1): #stupid range operator, it's gonna kill me
            # n<=k<=N
            Sum += self.ChooseTable[(N,k)] * pow(p,k) * pow((1-p),(N-k))
        return Sum

    def FillOutChooseTable(self, N):
        """Simply dose all the possible N choose k for k < N
        """
        self.FillOutFactorialTable(N)
        for k in range(N+1):
            if self.ChooseTable.has_key((N,k)):
                continue
            NFac = self.Factorial[N]
            kFac = self.Factorial[k]
            NMinusKFac = self.Factorial[N-k]
            Value = NFac / (kFac * NMinusKFac)
            self.ChooseTable[(N, k)] = Value

    def FillOutFactorialTable(self, X):
        """Get all the factorials of numbers <= X """
        for x in range(1, X+1):
            if self.Factorial.has_key(x):
                continue # already have this
            ##here we calculate it for the value x
            Prod = x
            for num in range(1, x):
                Prod *= num
            self.Factorial[x] = Prod
        #for Pair in self.Factorial.items():
        #    print "Factorials ",Pair

    def ComputePLS(self, N, nWinner, nRunnerUp):
        """From the Ascore Paper paper:
        AScore = ScoreWinner - ScoreRunnerUp
        Score = -10 * Log(P)
        P = Sum_{k=n}^N [N choose k] p^k * (1-p)^{N-k}
        k = iterator
        N = number of distinguishing ions
        n = number of distinguishing ions found
        p = some very poorly explained probability number.  It appears to be the probabilty of
            a peak in 100 m/z.  Inspect filters stuff automatically and I don't want to mess
            with it. so this number will be PeakDensity of 100 m/z units /100
        """
        self.FillOutChooseTable(N) # the N, or number of potential peaks
        TopBinomial = self.ComputeBinomial(N, nWinner)
        RunnerUpBinomial = self.ComputeBinomial(N, nRunnerUp)
        TopScore = -10 * math.log(TopBinomial, 10)
        RunnerUpScore = -10 * math.log(RunnerUpBinomial, 10)
        PLS = TopScore - RunnerUpScore # The AScore
        return PLS

    def GetDistinguishingPeaks(self, Peptide1, Peptide2):
        """ Given two peptides, find all the peaks that
        distinguish between the two phos placements, and return
        those in a list.  e.g. SphosEPTIDE vs. SEPTphosIDE
        distinguishing B fragments = b1: Sphos or S
                                    b2: SphosE or SE
                                    b3: SphosEP or SEP
                                    b4: SphosEPT and SEPTphos have the same mass = NOT DISTINGUISHING
    
        The general case is: if the b fragment has different number of phosphorylations
        between the two annotations, then both b and y are distinguishing.
        """
        Differences = [] #list of indicies that are different [1] would be differences on b1, yn-1
        ModIndex1 = Peptide1.Modifications.keys() # a list of indicies
        #print "Peptide %s"%Peptide1.GetModdedName()
        for Index in ModIndex1:
            #print "This is my mod on %s, %s"%(Index, Peptide1.Modifications[Index])
            PTMList = Peptide1.Modifications[Index]
            FoundPhos =0
            for Item in PTMList:
                if Item.Name == "Phosphorylation":
                    FoundPhos = 1
            if not FoundPhos:
                ModIndex1.remove(Index)
        ModIndex2 = Peptide2.Modifications.keys() # a list of indicies
        #print "Peptide %s"%Peptide2.GetModdedName()
        for Index in ModIndex2:
            #print "This is my mod on %s, %s"%(Index, Peptide2.Modifications[Index])
            PTMList = Peptide2.Modifications[Index]
            FoundPhos =0
            for Item in PTMList:
                if Item.Name == "Phosphorylation":
                    FoundPhos = 1
            if not FoundPhos:
                ModIndex2.remove(Index)
        Count1 =0
        Count2 =0
        for Index in range(len(Peptide1.Aminos)):
            if Index in ModIndex1:
                Count1 +=1
            if Index in ModIndex2:
                Count2 +=1
            if not Count1 == Count2:
                Differences.append(Index+1)
        ## now we have a list of the b indicies.  Let's make the return list
        DistinguishingPeaks = []
        for B in Differences:
            YIndex = len(Peptide1.Aminos) - B
            #print "a distinguishing peak B%s, Y%s"%(B, YIndex)
            DistinguishingPeaks.append("B%s"%B)
            DistinguishingPeaks.append("Y%s"%YIndex)
        return DistinguishingPeaks
            

    def GetAlternateAnnotations(self, Peptide):
        """Given an annotation(SAMPAYphosNE), return all
        alternate annotations.  This version should work correctly
        in the presence of non-phosphorylation modifications,
        e.g. SAM+16PAYphosNE
        """
        #Dummy = GetPeptideFromModdedName("SAM+16PAYphosNESphosT")
        #Peptide = Dummy
        NumPhos = Peptide.GetModdedName().count("phos")
        AllAnnotations = []
        if not NumPhos in [1,2]:
            return AllAnnotations # empty list
        ## now see if the number of phos == number of potential residues
        ## this means that an AScore is impossible, only one possible placement
        Count = Peptide.Aminos.count("S")
        Count += Peptide.Aminos.count("T")
        Count += Peptide.Aminos.count("Y")
        if Count == NumPhos:
            return AllAnnotations #key for "N/A"

        (Dephos, PhosPTM) = self.RemovePhosFromPeptide(Peptide)
        for Index in range(len(Dephos.Aminos)):
            if Dephos.Aminos[Index] in ["S", "T", "Y"]:
                #place a phosphorylation
                CreateNewLevel1 =0
                if not Dephos.Modifications.has_key(Index):
                    CreateNewLevel1 = 1
                    Dephos.Modifications[Index] = []
                Dephos.Modifications[Index].append(PhosPTM)
                #do I need to place a second phosphate?
                if NumPhos == 2:
                    for Jndex in range(Index + 1, len(Dephos.Aminos)):
                        if Dephos.Aminos[Jndex] in ["S", "T", "Y"]:
                            CreateNewLevel2 = 0
                            if not Dephos.Modifications.has_key(Jndex):
                                Dephos.Modifications[Jndex] = []
                                CreateNewLevel2 = 1
                            Dephos.Modifications[Jndex].append(PhosPTM)
                            #add string to list, remove phos and move on.
                            #print Dephos.GetModdedName()
                            AllAnnotations.append(Dephos.GetModdedName())
                            Dephos.Modifications[Jndex].pop()
                            if CreateNewLevel2:
                                del Dephos.Modifications[Jndex]
                else:
                    #only one phos.  Add string to list
                    AllAnnotations.append(Dephos.GetModdedName())
                    #print Dephos.GetModdedName()
                #only add string to the list if it's a single phos peptide
                #regardless, get rid of the modification now that we've done all it's combinations
                Dephos.Modifications[Index].pop() # get rid of the most recently added Phos PTM
                if CreateNewLevel1:
                    del Dephos.Modifications[Index]
        ## now one last thing remains.  We have to remove the original annotation from the list
        try:
            AllAnnotations.remove(Peptide.GetModdedName())
        except:
            print "The original peptide was not in the list.  that's BAAAAAAAADDDDDDDDDDDDDD"
            return []
        return AllAnnotations
                    

    def RemovePhosFromPeptide(self, Peptide):
        """Given a Peptide object, remove the phosphorylation PTMod
        objects and return the neutered copy
        """
        Clone = copy.deepcopy(Peptide)
        PTModPhos = None
        RemoveElement = [] #indicies in Clone.Modifications that become empty
        for AminoIndex in Clone.Modifications:
            ModificationList = Clone.Modifications[AminoIndex]
            for PTMod in ModificationList:
                if PTMod.Name == "Phosphorylation":
                    if not PTModPhos:
                        PTModPhos = PTMod
                    ModificationList.remove(PTMod)
            if len(ModificationList) == 0:
                RemoveElement.append(AminoIndex)
        ## now clone is without phosphorylations, but with other modifications
        ## clean up the empty Modification keys
        for Index in RemoveElement:
            del Clone.Modifications[Index]
        return (Clone, PTModPhos)

    def GetSupportingPeaks(self, FullPeakList, DistinguishingPeakList):
        SupportingPeaks = {} # key = b8 value = 1
        for Tuple in FullPeakList:
            Ion = Tuple[2]
            Index = Tuple[3]
            if Ion in ["B", "Y", "B2", "Y2"]:
                Peak = "%s%s"%(Ion[0], Index)
                if Peak in DistinguishingPeakList:
                    #print "Supporting peak found %s"%Peak
                    SupportingPeaks[Peak]= 1
        return len(SupportingPeaks)



#Title:          MakeImage.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
MakeImage.py takes a spectrum labeled with ion types (built by Label.py) and produces a graph.
"""
from Utils import *
import traceback
import MSSpectrum
import math
import traceback

try:
    from PIL import Image
    from PIL import ImageDraw
    from PIL import ImageFont
    # Fonts don't seem to work on Linux.  (Tried pdf, pcf, and pil formats...but no luck)
    # So, we'll content ourselves with a default font if we must:
    try:
        TheFont = ImageFont.truetype("Times.ttf", 12)
    except:
        TheFont = ImageFont.load_default()
except:
    traceback.print_exc()
    print "WARNING: Python Imaging Library (PIL) not installed.\n  Image creation is NOT available."
    Image = None


class WebColors:
    "Color scheme for web display.  Colorful."
    White = (255,255,255)
    Green = (0,255,0)
    Blue = (0,0,255)
    PaleBlue = (10,10,80)
    Red = (255,0,0)
    Grey = (155,155,155)
    #Grey = (0,0,0)
    Background = (255, 255, 255)
    Peak = (199, 199, 199)
    #Peak = (0,0,0)
    AnnotatedPeak = (0, 0, 0)
    Axis = (155, 155, 155)
    LabeledPeak = (0, 0, 0)
    PeakLabel = (200, 0, 0)
    BSeries = (55,55,200)
    BSeriesPale = (155,155,255)
    YSeries = (155,155,55)
    YSeriesPale = (200,200,100)

class PrintingColors:
    "Color scheme for printing.  Black-and-white, dark shades."
    White = (255,255,255)
    Green = (0,0,0)
    Blue = (0,0,0)
    PaleBlue = (80,80,80)
    Red = (0,0,0)
    Grey = (80,80,80)
    Background = (255, 255, 255)
    Peak = (200, 200, 200)
    AnnotatedPeak = (80, 80, 80)
    Axis = (80, 80, 80)
    LabeledPeak = (10, 10, 10)
    PeakLabel = (0, 0, 0)
    BSeries = (0,0,0)
    BSeriesPale = (180,180,180)
    YSeries = (0,0,0)
    YSeriesPale = (180,180,180)
    
#Colors = PrintingColors
Colors = WebColors

def SetColors(PrintFlag):
    global Colors
    if PrintFlag:
        Colors = PrintingColors
    else:
        Colors = WebColors
    print "COLORS SET!", Colors.YSeries
    
def RoundOff(Int):
    "Round an integer to the nearest power of ten"
    PowerOfTen = 10
    while (1):
        if PowerOfTen>Int:
            return Int
        Radix = Int % PowerOfTen
        if Radix == 0:
            PowerOfTen *= 10
            continue
        if Radix > PowerOfTen/2:
            return (Int-Radix) + PowerOfTen
        else:
            return (Int-Radix)
        
    
def GetTickWidth(MassWidth):
    "Look for a nice round number that divides the width into about 20 pieces."
    IdealSliceCount = 20
    # If 6-15, round to 10; if 26-35, round to 20...  If 151...
    #print "GetTickWidth:", MassWidth
    Width = int(round(MassWidth / IdealSliceCount))
    SliceCount = MassWidth / Width
    while (1):
        # Try rounding off another digit:
        RoundedWidth = RoundOff(Width)
        if RoundedWidth==Width:
            break # we can't round off any more!
        RoundedSliceCount = MassWidth / RoundedWidth
        if (RoundedSliceCount < IdealSliceCount / 2) or (RoundedSliceCount > IdealSliceCount * 2):
            break # Slices are too skinny or too fat
        Width = RoundedWidth
    return Width


class BaseImageMaker:
    """
    Graph generating class.  Not very MS-specific.  Subclassed by the spectrum plotter.
    """
    LeftPad = 30
    RightPad = 3
    UpperPad = 50
    LowerPad = 20
    def __init__(self, Width = 600, Height = 400):
        """
        Width is the total image width, in pixels.  The plot body ranges from
        self.LeftPad to self.Width.
        Height is the total image height, in pixels.  The plot body ranges from
        self.UpperPad to (self.Height-self.LowerPad)
        """
        self.MinX = 0
        self.MaxX = 100
        self.MinY = 0
        self.MaxY = 1
        self.Width = Width
        self.Height = Height
        self.BaseLine = self.Height - self.LowerPad
        self.YBreak = None
    def GetNiceTickMark(self, Width, GoodTickCount):
        """
        Given a width (axis size), compute a good interval for major tick marks.
        We want around 10 ticks, but we round the tick size up or down in order
        to get cleaner numbers.
        """
        if (Width <= 0):
            return 0.1
        Tick = Width / float(GoodTickCount)
        MinimumPower = int(math.log10(Tick)) - 1
        PossibleTicks = []
        for Power in range(MinimumPower, MinimumPower+5):
            PossibleTicks.append(10**Power)
            PossibleTicks.append(10**Power * 2)
            PossibleTicks.append(10**Power * 5)
        for Index in range(len(PossibleTicks) - 1):
            if PossibleTicks[Index+1] > Tick:
                return PossibleTicks[Index+1]
        return 0.1 # hacky fall-back case!
    def GetValueName(self, Value):
        """
        Take a number - potentially a very large or small one - and format it so as
        to use few characters, so as to make a usable axis label.
        1.2002 -> 1.2
        0.149 -> 0.15
        10000 -> 1e5
        -0.00005 -> -5e-5
        """
        if Value == 0:
            return "0"
        if Value < 0:
            Sign = "-"
            Value *= -1
        else:
            Sign = ""
        if Value > 1000:
            Exp = int(math.log10(Value))
            Abscissa = 10**(math.log10(Value) - int(math.log10(Value)))
            return "%s%.1fe%d"%(Sign,Abscissa, Exp)
        if Value < 0.01:
            Exp = math.floor(math.log10(Value))
            Abscissa = 10**(math.log10(Value) - Exp)
            return "%s%.1fe%d"%(Sign,Abscissa, Exp)
        if abs(Value) < 1:
            return "%s%.2f"%(Sign, Value)
        else:
            return "%s%.1f"%(Sign, Value)
    def DrawYAxis(self):
        if self.YBreak:
            self.DrawYAxisHelper(self.MinLowY, self.MaxLowY, 8)
            self.DrawYAxisHelper((self.MinHighY + self.MaxHighY) / 2.0, self.MaxHighY, 1)
            self.Draw.line((self.LeftPad, self.UpperPad, self.LeftPad, self.YBreak - 2), Colors.Axis)
            self.Draw.line((self.LeftPad - 4, self.YBreak + 4, self.LeftPad + 4, self.YBreak), Colors.Axis)
            self.Draw.line((self.LeftPad - 4, self.YBreak, self.LeftPad + 4, self.YBreak - 4), Colors.Axis)
            self.Draw.line((self.LeftPad, self.YBreak + 2, self.LeftPad, self.BaseLine), Colors.Axis)
        else:
            self.DrawYAxisHelper(self.MinY, self.MaxY, 10)
            self.Draw.line((self.LeftPad, self.UpperPad, self.LeftPad, self.BaseLine), Colors.Axis)
    def DrawYAxisHelper(self, MinVal, MaxVal, TickCount):
        "Draw the y-axis (including tick-marks and labels)"
        TickSize = self.GetNiceTickMark(MaxVal - MinVal, TickCount)
        ##print "Y axis: %s ticks from %s to %s; tick size %s"%(TickCount, MinVal, MaxVal, TickCount)
        MaxPowers = int(math.log10(MaxVal)) + 1
        LastY = None
        try:
            IntensityLevel = MinVal
            while IntensityLevel < MaxVal:
                Y = self.GetY(IntensityLevel)
                self.Draw.line((self.LeftPad-2,Y,self.LeftPad,Y),Colors.Axis)
                Str = self.GetValueName(IntensityLevel)
                self.Draw.text((0, Y-5), Str, Colors.Axis, font = TheFont)
                LastY = Y
                IntensityLevel += TickSize
        except:
            traceback.print_exc()
            print 0, self.MinY, self.MaxY, TickSize
            raise
    def BreakY(self, SmallestOfBig, BiggestOfSmall):
        self.MinLowY = self.MinY
        self.MaxLowY = BiggestOfSmall * 1.05
        self.MinHighY = max(self.MaxLowY * 1.001, SmallestOfBig * 0.95)
        self.MaxHighY = self.MaxY
        self.BrokenY = 1
        self.YBreak = int((self.Height - self.UpperPad - self.LowerPad) * 0.2)
        #print "Y break is:", self.YBreak
    def GetY(self, YValue):
        if self.YBreak:
            if YValue > self.MaxLowY:
                YPercent = (YValue - self.MinHighY) / max(1, self.MaxHighY - self.MinHighY)
                return self.YBreak - int((self.YBreak - self.UpperPad)*YPercent)
            else:
                YPercent = YValue / max(1, self.MaxLowY - self.MinLowY)
                return self.BaseLine - int((self.BaseLine - self.YBreak)*YPercent)
        YPercent = YValue / float(max(1, self.MaxY - self.MinY))
        return self.BaseLine - int((self.BaseLine - self.UpperPad) * YPercent)
    def GetX(self, XValue):
        XPercent = (XValue - self.MinX) / max(1, (self.MaxX - self.MinX))
        TotalWidth = self.Width - (self.LeftPad + self.RightPad)
        return self.LeftPad + int(XPercent * TotalWidth)
    def DrawTickMarks(self):
        TickPos = 0
        while TickPos < self.MaxX-1.0:
            if TickPos < self.MinX:
                TickPos += self.TickWidth
                continue
            X = self.GetX(TickPos)
            # Draw text, unless it would go over the edge:
            TextWidth = len(str(TickPos))*6
            TextX = X - TextWidth/2
            if TextX < self.Width - TextWidth:
                self.Draw.line((X, self.BaseLine, X, self.BaseLine+3), Colors.Axis)
                self.Draw.text((TextX, self.BaseLine+2), str(TickPos), Colors.Axis, font = TheFont)
            TickPos += self.TickWidth
    
class MSImageMaker(BaseImageMaker):
    def __init__(self, *args, **kw):
        # Some options:
        self.YBreakThreshold = 0.3333
        #self.Labels = {} Label -> (PeakX, PeakY, Label, PeakIntensity, IntensityRank)
        self.IntensityRank = {} # PeakIndex -> Rank
        BaseImageMaker.__init__(self, *args, **kw)
    def ConvertPeakAnnotationToImage(self, PeakAnnotationList, OutputFileName, Peptide = None, Width = 600, Height = 400):
        if not Image: # catch for no PIL
            return None
        self.Width = Width
        self.Height = Height
        self.BaseLine = self.Height - self.LowerPad
        self.PeakAnnotationList = PeakAnnotationList
        self.GetPeakDemographics(Peptide) #computes min, max, intensityrank
        self.PlotImage = Image.new("RGB", (Width, Height), Colors.Background)  # mode, size, [startcolor]
        self.Draw = ImageDraw.Draw(self.PlotImage)
        self.RoofLine = Height * 0.5
        MassWidth = self.Width - (self.LeftPad + self.RightPad)
        # Draw baseline
        self.Draw.line((self.LeftPad, self.BaseLine, Width - self.RightPad, self.BaseLine), Colors.Axis)
        # Draw x axis tickmarks (and labels):
        self.TickWidth = 200
        self.DrawTickMarks()
        # Draw y axis:
        self.DrawYAxis()
        # Draw peaks, with labels
        self.DrawPeaks()
        if Peptide:
            self.DrawBSeries(Peptide)
            self.DrawYSeries(Peptide)
        self.DrawPeakLabels()
        self.PlotImage.save(OutputFileName, "png")
    def GetPeakDemographics(self, Peptide = None):
        """
        Because this version uses a list of peak annotations given by
        PyInspect, and not peak objects, they don't come with an associated
        rank.  here I do a quick bubble sort and rank stuff (Ali, I wrote this bubble sort)
        """
        Intensities = []
        self.MinX = 1000
        self.MaxX = 0
        # Sort peak, from most to least intense:
        PeaksByIntensity = []
        for PeakIndex in range(len(self.PeakAnnotationList)):
            Tuple = self.PeakAnnotationList[PeakIndex]
            PeaksByIntensity.append((Tuple[1], Tuple[0], PeakIndex))
            ## set min and max masses
            if Tuple[0] > self.MaxX:
                self.MaxX = Tuple[0]
            if Tuple[0] < self.MinX:
                self.MinX = Tuple[0]
            self.MaxY = max(self.MaxY, Tuple[1])
        PeaksByIntensity.sort()
        PeaksByIntensity.reverse()
        ###
        self.IntensityRank = [None] * len(self.PeakAnnotationList)
        for IntensityRank in range(len(PeaksByIntensity)):
            self.IntensityRank[Tuple[-1]] = IntensityRank
        ## possibly reset x min and max
        if Peptide:
            self.MaxX = max(self.MaxX, Peptide.Masses[-1] + 10)
            self.MinX = 0
        FullMass = self.MaxX - self.MinX
        # Move the left and right edges a little bit, so that peaks don't
        # hit the edges of the image.
        self.MinX = max(0, self.MinX - FullMass * 0.05)
        self.MaxX = self.MaxX + FullMass * 0.05
        self.MaxY = max(10.0, self.MaxY)
        ##############
        ## do some Y breakage
        Intensity1 = PeaksByIntensity[0][0]
        if len(PeaksByIntensity) < 2:
            return # why we would have a spectrum with only one peak, I will never know.
        Intensity2 = PeaksByIntensity[1][0]
        if (Intensity2 / Intensity1 < self.YBreakThreshold):
            self.BreakY(Intensity1, Intensity2)
            return
        if len(PeaksByIntensity) < 3:
            return
        Intensity3 = PeaksByIntensity[2][0]
        if Intensity3 / Intensity2 < self.YBreakThreshold:
            self.BreakY(Intensity2, Intensity3)
            return
    def DrawPeaks(self):
        self.Labels = {}
        MaxIntensity = 0
        # First, no-ion-type peaks:
        for PeakTuple in self.PeakAnnotationList:
            (Mass, Intensity, Label, AminoIndex) = PeakTuple
            MaxIntensity = max(MaxIntensity, Intensity)
            if Label: # don't draw in grey anything which has been labeled
                continue
            PeakX = self.GetX(Mass)
            PeakY = self.GetY(Intensity)
            self.Draw.line((PeakX, PeakY, PeakX, self.BaseLine), Colors.Peak)
        MinLabelIntensity = MaxIntensity * 0.25 # any annotated peak above this threshold receives a text label, even if it's a neutral loss.
        # Next, peaks with assigned ions:
        for PeakIndex in range(len(self.PeakAnnotationList)):
            (Mass, Intensity, Label, AminoIndex) = self.PeakAnnotationList[PeakIndex]
            if not Label: #skip all the unlabeled peaks this round
                continue
            PeakX = self.GetX(Mass)
            PeakY = self.GetY(Intensity)
            self.Draw.line((PeakX, PeakY, PeakX, self.BaseLine), Colors.AnnotatedPeak)
            TextLabelPeakNames = ("B", "Y", "GlcNAc", "Y2", "B2")
            if Label in TextLabelPeakNames:  
                PeptideIndex = AminoIndex
                if Label in ("B", "Y"):
                    Label = Label.lower()
                if Label in ("B2", "Y2"):
                    Label = "%s2"%Label[0].lower() 
                if AminoIndex != None:
                    Label = "%s %s"%(Label, AminoIndex)
                OldLabelInfo = self.Labels.get(Label, None)
                if OldLabelInfo!=None and OldLabelInfo[-2] > Intensity:
                    continue
                self.Labels[Label] = (PeakX, PeakY, Label, Intensity, self.IntensityRank[PeakIndex])
            if Label[0] == "M":
                ##special case for phorphorylated spectra. the label
                ## Parent loss has been changed to M-p or M-p-h2o
                OldLabelInfo = self.Labels.get(Label, None)
                if OldLabelInfo != None and OldLabelInfo[-2] > Intensity:
                    continue
                self.Labels[Label] = (PeakX, PeakY, Label, Intensity, self.IntensityRank[PeakIndex])
    def CollideRectangles(self, X1, Y1, X2, Y2, Rectangles):
        for (CX1, CY1, CX2, CY2) in Rectangles:
            if CX1 <= X1 <= CX2 and CY1 <= Y1 <= CY2:
                return (CX1, CY1, CX2, CY2)
            if CX1 <= X2 <= CX2 and CY1 <= Y1 <= CY2:
                return (CX1, CY1, CX2, CY2)
            if CX1 <= X1 <= CX2 and CY1 <= Y2 <= CY2:
                return (CX1, CY1, CX2, CY2)
            if CX1 <= X2 <= CX2 and CY1 <= Y2 <= CY2:
                return (CX1, CY1, CX2, CY2)
        return None
    def DrawPeakLabels(self):
        # Sort labels by priority.  b and y take precedence over all else;
        # intense peaks take precedence over others.
        SortedLabels = []
        for (X, Y, Label, Intensity, Rank) in self.Labels.values():
            if Rank < 10:
                Priority = 0
            else:
                Priority = 1
            SortedLabels.append([Priority, Y, X, Label, Intensity, None, None])
        SortedLabels.sort()
        SortedLabels = SortedLabels[:25]
        DirtyRectangles = []
        for List in SortedLabels:
            (IsBY, Y, X, Label, Intensity, Dummy, Dummy) = List
            (Width, Height) = self.Draw.textsize(Label, font = TheFont)
            Height *= 2 # for superscript and subscript
            X1 = X - Width/2
            X2 = X + Width/2
            Y1 = Y - Height
            Y2 = Y
            Tuple = self.CollideRectangles(X1, Y1, X2, Y2, DirtyRectangles)
            if Tuple == None:
                List[5] = (X1, Y1, X2, Y2)
                DirtyRectangles.append((X1, Y1, X2, Y2))
                continue
            (CX1, CY1, CX2, CY2) = Tuple
            # Try moving this label off to the side:
            if (X1 + X2) / 2 < (CX1 + CX2) / 2:
                Move = (X2 - CX1) + 1
                X1 -= Move
                X2 -= Move
                Y1 -= 5
                Y2 -= 5
            else:
                Move = (CX2 - X1) + 1
                X1 += Move
                X2 += Move
                Y1 -= 5
                Y2 -= 5
            Tuple = self.CollideRectangles(X1, Y1, X2, Y2, DirtyRectangles)
            if Tuple == None:
                List[5] = (X1, Y1, X2, Y2)
                DirtyRectangles.append((X1, Y1, X2, Y2))
                List[6] = ((X1 + X2) / 2, Y2, X, Y)
                continue
        for Index in range(len(SortedLabels)-1, -1, -1):
            List = SortedLabels[Index]
            if List[5]:
                (X1, Y1, X2, Y2) = List[5]
                self.Draw.line((List[2], List[1], List[2], self.BaseLine), Colors.LabeledPeak) # color the peak
        for Index in range(len(SortedLabels)-1, -1, -1):
            List = SortedLabels[Index]
            if List[5]:
                (X1, Y1, X2, Y2) = List[5]
                # Most peaks are drawn using superscripts and subscripts:
                PeakName = List[3]
                if PeakName == "GlcNAc":
                    self.Draw.text((X1, Y1 + 5), PeakName, Colors.PeakLabel, font = TheFont)
                else:
                    self.Draw.text((X1, Y1 + 5), PeakName[0], Colors.PeakLabel, font = TheFont)
                    NumIndex = len(PeakName) - 1
                    while PeakName[NumIndex] in ("0123456789"):
                        NumIndex -= 1
                    if NumIndex > 0:
                        SuperScript = PeakName[1:NumIndex+1].strip()
                    else:
                        SuperScript = ""
                    if SuperScript:
                        self.Draw.text((X1+7, Y1+10), SuperScript, Colors.PeakLabel, font = TheFont)
                        self.Draw.text((X1+7, Y1), PeakName[NumIndex+1:], Colors.PeakLabel, font = TheFont)
                    else:
                        self.Draw.text((X1+7, Y1+5), PeakName[NumIndex+1:], Colors.PeakLabel, font = TheFont)
            # Draw the dotted line from the label to its peak:                
            if List[6]:
                self.Draw.line(List[6], Colors.LabeledPeak)
    def DrawDottedLine(self, X1, Y1, X2, Y2, Color):
        Distance = math.sqrt((X2-X1)**2 + (Y2-Y1)**2)
        if Distance == 0:
            return
        DX = (X2-X1)/Distance
        DY = (Y2-Y1)/Distance
        OldLineLength = 0
        Dot = 1
        while (1):
            LineLength = min(Distance, OldLineLength + 5)
            XA = int(X1 + DX*OldLineLength)
            XB = int(X1 + DX*LineLength)
            YA = int(Y1 + DY*OldLineLength)
            YB = int(Y1 + DY*LineLength)
            if Dot:
                self.Draw.line((XA, YA, XB, YB), Color)
            OldLineLength = LineLength
            Dot = not Dot
            if (LineLength == Distance):
                break
    def DrawBSeries(self, Peptide):
        BHeight = 17
        if getattr(Peptide, "Seed", None):
            SeedIndex = Peptide.Aminos.rfind(Peptide.Seed)
        else:
            SeedIndex = -999
        self.Draw.text((10, BHeight-7), "b", Colors.BSeries, font = TheFont)
        # First, draw tickmarks for the b peaks
        for MassIndex in range(len(Peptide.Masses)):
            Mass = Peptide.Masses[MassIndex]
            X = self.GetX(Mass)
            Label = ("b %s"%MassIndex)
            if self.Labels.has_key(Label):
                PeakIntensity = self.Labels[Label][-2]
                Y = self.GetY(PeakIntensity)
                self.DrawDottedLine(X, BHeight-2, X, Y, Colors.BSeriesPale)
            else:
                self.Draw.line((X, BHeight-2, X, BHeight+2), Colors.BSeriesPale)
        # Now draw horizontal lines, and amino labels:
        for AminoIndex in range(len(Peptide.Aminos)):
            LabelA = "b %s"%AminoIndex
            LabelB = "b %s"%(AminoIndex+1)
            XA = self.GetX(Peptide.Masses[AminoIndex])
            XB = self.GetX(Peptide.Masses[AminoIndex+1])
            HasLabelA = self.Labels.has_key(LabelA)
            if AminoIndex == 0:
                HasLabelA = 1
            HasLabelB = self.Labels.has_key(LabelB)
            if AminoIndex ==len(Peptide.Aminos)-1:
                HasLabelB = 1
            if HasLabelA and HasLabelB:
                self.Draw.line((XA, BHeight, XB, BHeight), Colors.BSeries)
            else:
                self.DrawDottedLine(XA, BHeight, XB, BHeight, Colors.BSeriesPale)
            if AminoIndex in (SeedIndex, SeedIndex+1, SeedIndex+2):
                self.Draw.line((XA, BHeight-1, XB, BHeight-1), Colors.BSeries)
                self.Draw.line((XA, BHeight, XB, BHeight), Colors.BSeries)
                self.Draw.line((XA, BHeight+1, XB, BHeight+1), Colors.BSeries)
                
            X = (XA+XB)/2 - 3
            Str = Peptide.Aminos[AminoIndex]
            if Peptide.Modifications.get(AminoIndex):
                self.Draw.text((X-4, BHeight-14), "%s*"%Str, Colors.BSeries, font = TheFont)
            else:
                self.Draw.text((X, BHeight-14), Str, Colors.BSeries, font = TheFont)
    def DrawYSeries(self, Peptide): # Copied and modded from DrawBSeries
        YHeight = 34
        if getattr(Peptide, "Seed", None):
            SeedIndex = Peptide.Aminos.rfind(Peptide.Seed)
        else:
            SeedIndex = -999
        
        self.Draw.text((10, YHeight-7), "y", Colors.YSeries, font = TheFont)
        # First, draw tickmarks for the y peaks
        PM = Peptide.Masses[-1] + 19 # parent mass
        for MassIndex in range(len(Peptide.Masses)):
            Mass = PM - Peptide.Masses[MassIndex]
            X = self.GetX(Mass)
            Label = "y %s"%(len(Peptide.Masses) - MassIndex - 1)
            #Label = "y %s"%(MassIndex)
            #print "Y series y %d, mass %f, label %s"%(MassIndex, Mass, Label)
            if self.Labels.has_key(Label):
                PeakIntensity = self.Labels[Label][-2]
                Y = self.GetY(PeakIntensity)
                self.DrawDottedLine(X, YHeight-2, X, Y, Colors.YSeriesPale)
            else:
                self.Draw.line((X, YHeight-2, X, YHeight+2), Colors.YSeriesPale)
        # Now draw horizontal lines, and amino labels:
        for AminoIndex in range(len(Peptide.Aminos)):
            LabelA = "y %s"%(len(Peptide.Aminos) - AminoIndex)
            LabelB = "y %s"%(len(Peptide.Aminos) - AminoIndex - 1)
            XA = self.GetX(PM - Peptide.Masses[AminoIndex])
            XB = self.GetX(PM - Peptide.Masses[AminoIndex+1])
            HasLabelA = self.Labels.has_key(LabelA)
            if AminoIndex == 0:
                HasLabelA = 1
            HasLabelB = self.Labels.has_key(LabelB)
            if AminoIndex ==len(Peptide.Aminos)-1:
                HasLabelB = 1
            if HasLabelA and HasLabelB:
                self.Draw.line((XA, YHeight, XB, YHeight), Colors.YSeries)
            else:
                self.DrawDottedLine(XA, YHeight, XB, YHeight, Colors.YSeriesPale)
            if AminoIndex in (SeedIndex, SeedIndex+1, SeedIndex+2):
                self.Draw.line((XA, YHeight-1, XB, YHeight-1), Colors.YSeries)
                self.Draw.line((XA, YHeight, XB, YHeight), Colors.YSeries)
                self.Draw.line((XA, YHeight+1, XB, YHeight+1), Colors.YSeries)
                
            X = (XA+XB)/2 - 3
            Str = Peptide.Aminos[AminoIndex]
            if Peptide.Modifications.get(AminoIndex):
                self.Draw.text((X-4, YHeight-14), "%s*"%Str, Colors.YSeries, font = TheFont)
            else:
                self.Draw.text((X, YHeight-14), Str, Colors.YSeries, font = TheFont)

UsageInfo = """
Usage:
   MakeImage.py <LabeledSpectrum> [<OutputFileName>]
"""

if __name__ == "__main__":
    if len(sys.argv)<2:
        print UsageInfo
        sys.exit(1)
    InputFileName = sys.argv[1]
    if len(sys.argv)>2:
        OutputFileName = sys.argv[2]
    else:
        OutputFileName = os.path.splitext(InputFileName)[0] + ".png"
    Maker = MSImageMaker()
    Maker.ConvertSpectrumFileToImage(InputFileName, OutputFileName)



#Title:          PrepDB.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
Translate a protein database to a good format for trie-based searching.
The source database should be in either FASTA format or in swiss-prot format.
The output database will be in "concatenated format" - peptide strings with
asterisks delimiting the peptides, no whitespace.
We also save a binary file indexing into the concatenated DB.

Index file format is record-based, with one record per peptide:
- original DB position (int); the START of a record (>)
- concatenated DB file position (int); the START of a record (first peptide)
- Peptide ID (string, 80 chars)
"""
import sys
import struct
import traceback
import os
import string

class SwissCompressor:
    """
    Convert a protein database into concatenated format.
    Processes the SwissProt database format.
    """
    def __init__(self, SourceFileName, SquishedFileName, IndexFileName, Species = None):
        self.SourceFile = open(SourceFileName,"rb")
        self.SquishedFile = open(SquishedFileName,"wb")
        self.IndexFile = open(IndexFileName,"wb")
        self.FASTA = 0
        self.Species = Species
    def Compress(self):
        """
        The parts of swiss-prot we care about look like this:
SQ   SEQUENCE   296 AA;  34077 MW;  B0D7CD175C7A3625 CRC64;
     FNSNMLRGSV CEEDVSLMTS IDNMIEEIDF YEKEIYKGSH SGGVIKGMDY DLEDDENDED
     EMTEQMVEEV ADHITQDMID EVAHHVLDNI THDMAHMEEI VHGLSGDVTQ IKEIVQKVNV
     AVEKVKHIVE TEETQKTVEP EQIEETQNTV EPEQTEETQK TVEPEQTEET QNTVEPEQIE
     ETQKTVEPEQ TEEAQKTVEP EQTEETQKTV EPEQTEETQK TVEPEQTEET QKTVEPEQTE
     ETQKTVEPEQ TEETQKTVEP EQTEETQKTV EPEQTEETQN TVEPEPTQET QNTVEP
//
        """        
        self.InSequence = 0
        RecordNumber = 0
        LineNumber = 0
        CorrectSpecies = 0
        while (1):
            LineNumber += 1
            SourceFilePos = self.SourceFile.tell()
            RawFileLine  = self.SourceFile.readline()
            if not RawFileLine:
                break # end o' file!
            FileLine = RawFileLine.strip()
            if self.InSequence:
                # // marks end of sequence; anything else is sequence data.
                # ...but in some cases, the // marker isn't present, so we
                # stop when we see the "ID" tag from the next record.
                #if FileLine[:2] == "//":
                #print self.InSequence, FileLine
                if RawFileLine[:2] != "  ":
                    self.InSequence = 0
                    if self.FASTA:
                        pass
                    else:
                        self.SquishedFile.write("*")
                    RecordNumber += 1
                else:
                    Stripped = FileLine.replace(" ","")
                    self.SquishedFile.write(Stripped)
            else:
                if FileLine[:3] == "OS ":
                    if self.Species == None or FileLine.lower().find(self.Species)!=-1:
                        CorrectSpecies = 1
                    else:
                        CorrectSpecies = 0
                if FileLine[:3] == "ID ":
                    SourceFileRecordStart = SourceFilePos
                    ID = FileLine.split()[1]
                    ID = ID[:80]
                    if self.FASTA:
                        self.SquishedFile.write("\n>%s\n"%ID)
                if FileLine[:3] == "SQ ":
                    if CorrectSpecies:
                        self.InSequence = 1
                        SquishedFilePos = self.SquishedFile.tell()
                        Str = struct.pack("<qi80s", SourceFileRecordStart, SquishedFilePos, ID)
                        self.IndexFile.write(Str)
            if LineNumber%1000 == 0:
                print "Processed line %d."%LineNumber
                #self.SquishedFile.flush()
                #self.IndexFile.flush()
                #sys.stdin.readline()
        print "Total records seen:", RecordNumber

class FASTACompressor:
    """
    Convert a protein database into concatenated format.
    Processes FASTA format.
    """
    def __init__(self, SourceFileName, SquishedFileName, IndexFileName, Species = None):
        self.SourceFile = open(SourceFileName,"rb")
        self.SquishedFile = open(SquishedFileName,"wb")
        self.IndexFile = open(IndexFileName,"wb")
        self.SquishedFileName = SquishedFileName
        self.IndexFileName = IndexFileName
    def Compress(self):
        RecordNumber = 0
        LineNumber = 0
        FirstRecord = 1
        LineNumberWarnings = 0
        DummyTable = string.maketrans("", "")
        while (1):
            LineNumber += 1
            SourceFilePos = self.SourceFile.tell()
            FileLine  = self.SourceFile.readline()
            if not FileLine:
                break # end o' file!
            FileLine = FileLine.strip()
            if not FileLine:
                continue # empty lines (whitespace only) are skipped
            if FileLine[0] == ">":
                RecordNumber += 1
                if not FirstRecord:
                    self.SquishedFile.write("*")                
                ID = FileLine[1:81].strip()
                # Fix weird characters in the ID:
                ID = ID.replace("\t", " ")
                # Note: Important to call tell() *after* writing the asterisk!  (Fixed a bug 1/20/5)
                SquishedFilePos = self.SquishedFile.tell() 
                Str = struct.pack("<qi80s", SourceFilePos, SquishedFilePos, ID)
                self.IndexFile.write(Str)
                FirstRecord = 0
            else:
                WarnFlag = 0
                FileLine = string.translate(FileLine, DummyTable, " \r\n\t*")
                FileLine = FileLine.upper()
                Str = ""
                for Char in FileLine:
                    if Char not in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                        WarnFlag = 1
                    else:
                        Str += Char
                #FileLine = FileLine.replace("*","")
                if WarnFlag and LineNumberWarnings < 10:
                    print "* Warning: line %s contains non-amino-acid characters:"%LineNumber
                    print FileLine
                    LineNumberWarnings += 1
                    if LineNumberWarnings >= 10:
                        print "(omitting further warnings)"
                self.SquishedFile.write(Str)
        print "Converted %s protein sequences (%s lines) to .trie format."%(RecordNumber + 1, LineNumber)
        print "Created database file '%s'"%self.SquishedFileName

class MS2DBCompressor:
    """
    Creates the index file for a splice graph, no modification is made to the original database
    """
    def __init__(self, SourceFileName, SquishedFileName, IndexFileName, Species = None):
        self.SourceFile = open(SourceFileName,"rb")
        
        self.IndexFile = open(IndexFileName,"wb")
        self.IndexFileName = IndexFileName
    def Compress(self):
        RecordNumber = 0
        LineNumber = 0
        FirstRecord = 1
        LineNumberWarnings = 0
        DummyTable = string.maketrans("", "")
        while (1):
            LineNumber += 1
            SourceFilePos = self.SourceFile.tell()
            FileLine  = self.SourceFile.readline()
            if not FileLine:
                break # end o' file!
            FileLine = FileLine.strip()
            if not FileLine:
                continue # empty lines (whitespace only) are skipped
            if FileLine[0:6] == "<Gene ":
                RecordNumber += 1
                
                ID = ""
                Bits = FileLine[6:].split(" ")
                for B in Bits:
                    (Item,Value) = B.split("=")
                    if Item == "Name":
                        ID = Value[1:-1]
                        break

                if ID == "":
                    print "No valid ID found in %s"%FileLine
                    raw_input()
                
                # Note: Important to call tell() *after* writing the asterisk!  (Fixed a bug 1/20/5)
                Str = struct.pack("<qi80s", SourceFilePos, SourceFilePos, ID)
                self.IndexFile.write(Str)
                FirstRecord = 0
            
        print "Converted %s protein sequences (%s lines) to .ms2index format."%(RecordNumber + 1, LineNumber)
        print "Created index file '%s'"%self.IndexFileName


def PrintUsage():
    print "Please supply a database filename."
    print "Usage: PrepDB.py <format> <OriginalDB> [NewDB] [IndexFile]"
    print "Example: Prepdb.py FASTA Drome.fasta"
    print "  The source format can be either FASTA or SWISS or MS2DB"
    print "  New DB file name defaults to original filename with .trie appended (no new file is created for MS2DB)"
    print "  Index file name defaults to original filename with .index appended (or .ms2index for MS2DB)"

if __name__ == "__main__":    
    if len(sys.argv)<3:
        PrintUsage()
        sys.exit()
    try:
        import psyco
        psyco.full()
    except:
        print "(psyco not found - running in non-optimized mode)"
    # First argument: Original database file format
    Format = sys.argv[1].lower()
    if Format == "fasta":
        CompressorClass = FASTACompressor
    elif Format == "swiss":
        CompressorClass = SwissCompressor
    elif Format == "ms2db":
        CompressorClass = MS2DBCompressor
    else:
        print "Unknown source database format '%s'"%Format
        PrintUsage()
        sys.exit()
    # Second argument: Original database file
    SourceFileName = sys.argv[2]
    # Optional third argument: New database file name
    if len(sys.argv) > 3:
        SquishedFileName = sys.argv[3]
    elif Format == "ms2db":
        SquishedFileName = None
    else:
        SquishedFileName = "%s.trie"%os.path.splitext(SourceFileName)[0]
    # Optional third argument: Index file name
    if len(sys.argv) > 4:
        IndexFileName = sys.argv[4]
    elif Format == "ms2db":
        IndexFileName = "%s.ms2index"%os.path.splitext(SourceFileName)[0]
    else:
        IndexFileName = "%s.index"%os.path.splitext(SourceFileName)[0]
    # Use FASTACompressor for FASTA format, Compressor for the weird swiss-prot format
    # If "species" is a string, then the Swiss-prot reader will filter out any records
    # that don't contain that string.  For example, set Species = "sapiens" to grab only
    # human proteins.
    Species = None
    Squasher = CompressorClass(SourceFileName, SquishedFileName, IndexFileName, Species)
    Squasher.Compress()



#Title:          BuildMGF.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.

"""
BuildMGF.py
This is a part of the PTMAnalysis pipeline.  It creates an
.mgf file of all the consensus spectra created by ComputePTMFeatures.py
"""

import os
import sys
import string
import getopt
import MSSpectrum
from Utils import *
Initialize()
from TrainPTMFeatures import FormatBits

class PeptideFeatureBag:
    pass

class MGFBuilder:
    def __init__(self):
        self.ConsensusSpectrumDir = "ptmscore\\LensLTQ-99-5\\spectra"
        self.PeptideFeatureFileName = "PTMScore\\LensLTQ-99-5.txt"
        self.MGFPath = "PTMScore\\LensLTQ-99-5.mgf"
        self.ModifiedPeptides = []
    def ParsePeptideFeatureFile(self):
        """
        Parse the contents of the peptide feature-file.  We need to know the
        path to the consensus spectrum file, the consensus annotation MQScore,
        and the index.
        """
        File = open(self.PeptideFeatureFileName, "rb")
        LineNumber = 0
        for FileLine in File.xreadlines():
            LineNumber +=1
            if FileLine[0] == "#":
                continue
            Bits = list(FileLine.replace("\r", "").replace("\n", "").split("\t"))
            try:
                ConsensusMQScore = float(Bits[FormatBits.ConsensusMQScore])
            except:
                print "** Error: Can't parse consensus MQScore from line %s!"%LineNumber
                print Bits
                continue
            PeptideFeatures = PeptideFeatureBag()
            PeptideFeatures.Bits = Bits
            PeptideFeatures.ConsensusMQScore = ConsensusMQScore
            NiceAnnotation = Bits[FormatBits.Peptide].replace("*", "-")
            PeptideFeatures.Bits[FormatBits.Peptide] = NiceAnnotation
            FirstResidue = NiceAnnotation[2]
            Charge = Bits[FormatBits.Charge]
            PeptideFeatures.SpectrumPath = os.path.join(self.ConsensusSpectrumDir, FirstResidue, "%s.%s.dta"%(NiceAnnotation, Charge))
            self.ModifiedPeptides.append(PeptideFeatures)
        File.close()
        print "Parsed %s modified peptides from %s file lines."%(len(self.ModifiedPeptides), LineNumber)
    def PrepareSearchMGF(self):
        """
        Concatenate our consensus spectra into an MGF file, for searching.
        """
        MGFFile = open(self.MGFPath, "wb")
        ScanNumber = 1
        for PeptideIndex in range(len(self.ModifiedPeptides)):
            if PeptideIndex % 100 == 0:
                print "Peptide species %s/%s..."%(PeptideIndex, len(self.ModifiedPeptides))
            PeptideFeatures = self.ModifiedPeptides[PeptideIndex]
            Spectrum = MSSpectrum.SpectrumClass()
            Spectrum.ReadPeaks(PeptideFeatures.SpectrumPath)
            Spectrum.WriteMGFPeaks(MGFFile, PeptideFeatures.Bits[FormatBits.Peptide], ScanNumber)
            ScanNumber += 1
        MGFFile.close()
    def Main(self):
        self.ParsePeptideFeatureFile()
        self.PrepareSearchMGF()
    def ParseCommandLine(self, Arguments):
        (Options, Args) = getopt.getopt(Arguments, "d:m:")
        OptionsSeen = {}
        for (Option, Value) in Options:
            OptionsSeen[Option] = 1
            if Option == "-d":
                self.PeptideFeatureDir = Value
            elif Option == "-m":
                self.MGFPath = Value
        self.ConsensusSpectrumDir = os.path.join(self.PeptideFeatureDir, "Spectra")
        self.PeptideFeatureFileName = os.path.join(self.PeptideFeatureDir, "PTMFeatures.txt")

UsageInfo = """
BuildMGF arguments:
  -d [DIR]: Peptide feature directory
  -m [FILE]: Output .mgf file name
"""
  
if __name__ == "__main__":
    Builder = MGFBuilder()
    Builder.ParseCommandLine(sys.argv[1:])
    Builder.Main()



#Title:          MSSpectrum.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""
Classes representing an MS/MS spectrum and its peaks.
"""
import traceback
import sys
import os
import types
import string
import re
import struct
import math
import base64 # for mzxml parsing
import ParseXML
from Utils import *
Initialize()

# Some systems have old versions of base64, without the new decode interface.
# As a workaround, set "B64Decode" to the decoding function.
if hasattr(base64, "b64decode"):
    B64Decode = base64.b64decode
    B64Encode = base64.b64encode
else:
    B64Decode = base64.decodestring
    B64Encode = base64.encodestring

class PeakClass:
    """
    One peak from an ms/ms spectrum.  Mostly just a mass, but we track the intensity and
    (perhaps putative) ion-type, as well.  Note that this may be a spectral peak (in
    which case the mass is a spectral mass) or a PRM peak (in which case the mass is
    a prefix residue mass).  Each spectral peak gives rise to n PRM peaks, where n
    is the number of ion types available.
    """
    def __init__(self, Mass, Intensity):
        self.Mass = Mass
        self.Intensity = Intensity
        self.IonType = None # Assigned only for PRM peaks
        self.Score = 0 # Score, based on current filtering scheme.
        self.PeptideIndex = None
        self.FilterRank = None
        self.IntensityRank = None 
        self.IsPlausibleIsotopicPeak = 0
        self.HasPlausibleIsotopicPeak = 0
    def __cmp__(self, Other):
        "Sort two peak objects - compare the masses"
        if not isinstance(Other, PeakClass):
            return 1
        if (self.Mass < Other.Mass):
            return -1
        if (self.Mass > Other.Mass):
            return 1
        return 0
    def __str__(self):
        return "<peak %s>"%self.Mass
    def GetPeakMass(self, ParentMass):
        "Get the corresponding spectral mass from this PRM peak"
        return self.IonType.GetPeakMass(self.Mass, ParentMass)
    def GetPRMMass(self, ParentMass):
        "Get the corresponding PRM mass from this spectral peak"
        return self.IonType.GetPRMMass(self.Mass, ParentMass)
    def PrintMe(self):
        print "Printing information for a PeakClass object"
        print "Mass %f Intensity %f "%(self.Mass,self.Intensity)
        print "IonType %s PeptideIndex %s IntensityRank %d"%(self.IonType, self.PeptideIndex, self.IntensityRank)

class SpectrumClass:
    """
    Mass-spec data, and some functions to filter peaks and such.  
    """
    InstanceCount = 0
    def __del__(self):
        SpectrumClass.InstanceCount -= 1
    def __init__(self, Scoring = None):
        SpectrumClass.InstanceCount += 1
        # Init our attributes:
        self.Name = None
        self.ParentMass = None
        self.PrecursorMZ = None
        self.PrecursorIntensity = None
        self.Charge = 1 # default
        self.Peaks = None # list of PeakClass instances
        self.PRMPeaks = None # list of PeakClass instances
        # The actual parent peptide (instance of PeptideClass), if known:
        self.CorrectPeptide = None
        self.Scoring = Scoring
    def GetSignalToNoise(self):
        "Return signal-to-noise ratio for this spectrum"
        Intensities = []
        for Peak in self.Peaks:
            Intensities.append(Peak.Intensity)
        Intensities.sort()
        IntenseCount = min(len(Intensities), 5)
        if not IntenseCount:
            return 0
        Signal = Intensities[-IntenseCount/2]
        Noise = Intensities[len(Intensities)/2]
        return Signal / float(Noise)
    def GetTotalIntensity(self):
        Intensity = 0
        for Peak in self.Peaks:
            Intensity += Peak.Intensity
        return Intensity 
    def SetCharge(self, NewCharge):
        self.Charge = NewCharge
        self.ParentMass = self.PrecursorMZ * NewCharge - (NewCharge - 1)*1.0078
    def ReadPeaksMGF(self, File):
        self.Peaks = []
        for FileLine in File.xreadlines():
            if FileLine[:8] == "END IONS":
                break
            if FileLine[:6] == "CHARGE":
                # Note: "2+ and 3+" is NOT supported.  Use the MultiCharge option,
                # or include two scans in the input file.
                Charge = int(FileLine[7:9].replace("+",""))
                continue
            if FileLine[:7] == "PEPMASS":
                #self.ParentMass = float(FileLine[8:])
                self.PrecursorMZ = float(FileLine[8:].split()[0])
                continue
            Bits = FileLine.split()
            try:
                Mass = float(Bits[0])
                Intensity = float(Bits[1])
            except:
                continue # some other header line we don't eat.
            Peak = PeakClass(Mass, Intensity)
            self.Peaks.append(Peak)
        if self.Charge == 0:
            # Guess!
            self.Charge = 2
        self.ParentMass = (self.PrecursorMZ * self.Charge) - (self.Charge-1)*1.0078
        #print "PrecursorMZ %s charge %s"%(self.PrecursorMZ, self.Charge)
    def ReadPeaksDTA(self, File):
        "Read a spectrum from a file, assuming .dta or .pkl or .ms2 format."
        HeaderLine = File.readline()
        if not HeaderLine.strip():
            HeaderLine = File.readline()
            if not HeaderLine.strip():
                HeaderLine = File.readline()
                if not HeaderLine.strip():
                    HeaderLine = File.readline()
                    if not HeaderLine.strip():
                        HeaderLine = File.readline()
        #print "HeaderLine: '%s'"%HeaderLine.strip()
        Bits = HeaderLine.strip().split()
        if HeaderLine[:7]=="CHARGE=":
            self.Charge = int(HeaderLine[7])
            HeaderLine = File.readline()
            self.PrecursorMZ = float(HeaderLine[8:])
            self.ParentMass = (self.PrecursorMZ * self.Charge) - (self.Charge-1)*1.0078
        elif Bits[0] == "Z": # MS2 format:
            self.Charge = int(Bits[1])
            self.ParentMass = float(Bits[2])
        elif Bits[0] == "S": #MS2 format:
            HeaderLine = File.readline()
            Bits = HeaderLine.strip().split()
            if Bits[0] == "Z":
              self.Charge = int(Bits[1])
              self.ParentMass = float(Bits[2])  
            else:
                print "ERROR: Expecting a line starting with Z but instead found %s"%HeaderLine
                sys.exit(0)
        elif HeaderLine[0] == ":": # MS2 colon format:
            HeaderLine = File.readline()
            Bits = HeaderLine.strip().split()
            self.ParentMass = float(Bits[0])
            self.Charge = int(Bits[1]) # always an integer!
        elif len(Bits) == 3: #PKL format:
            self.PrecursorMZ = float(Bits[0])
            self.Charge = int(Bits[2])
            self.ParentMass = (self.PrecursorMZ * self.Charge) - (self.Charge-1)*1.0078
        else:
            self.ParentMass = float(Bits[0])
            self.Charge = int(Bits[1]) # always an integer!
        if self.Charge == 0:
            # Guess!
            self.Charge = 2
            self.PrecursorMZ = self.ParentMass
            self.ParentMass = (self.PrecursorMZ * self.Charge) - (self.Charge-1)*1.0078
            #print "Prescursor MZ is %.2f, so guess a parent mass of %.2f"%(self.PrecursorMZ, self.ParentMass)
        else:
            self.PrecursorMZ = (self.ParentMass + (self.Charge-1)*1.0078) / self.Charge
        self.Peaks = []
        for FileLine in File.xreadlines():
            Bits = FileLine.split()
            # Skip comments:
            if FileLine[0] == "#":
                continue
            Bits = FileLine.split()
            if not Bits:
                break
            if Bits[0] == "Z":
                continue # special for ms2: ignore.
            if len(Bits) > 2:
                break # no more!
            try:
                Mass = float(Bits[0])
                Intensity = float(Bits[1])
            except:
                # It's over, over, over.
                break
            Peak = PeakClass(Mass, Intensity)
            self.Peaks.append(Peak)
        #File.close()
        self.Peaks.sort() # sort by mass

    def ReadPeakDTALine(self, FileLine):
        FileLine = FileLine.strip()
        Bits = FileLine.split()
        if len(Bits) < 2:
            return # blank (or broken) line, skip
        Peak = PeakClass(float(Bits[0]), float(Bits[1]))
        # If this is a labeled .dta file, read the ion types and the peptide indices:
        if len(Bits)>2:
            Peak.IonType = Global.AllIonDict.get(Bits[2], None)
        if len(Bits)>3:
            try:
                Peak.PeptideIndex = int(Bits[3])
            except:
                pass # silent failure (a novel)
        self.Peaks.append(Peak)
    def RankPeaksByIntensity(self):
        "Set Peak.IntensityRank for each of our peaks."
        PeaksSortedByIntensity = []
        for Peak in self.Peaks:
            PeaksSortedByIntensity.append((Peak.Intensity, Peak))
        PeaksSortedByIntensity.sort()
        PeaksSortedByIntensity.reverse()
        for Index in range(len(PeaksSortedByIntensity)):
            PeaksSortedByIntensity[Index][1].IntensityRank = Index
    def ReadPeaksPKL(self, File):
        "Read peaks from a file in .pkl format"
        HeaderLine = File.readline()
        Bits = HeaderLine.split()
        if len(Bits)!=3:
            # .pkl files should have precursor m/z, precursor peak intensity, and
            # guessed charge.  If we don't have three pieces, then this isn't a
            # valid .pkl file...
            raise ValueError, "Invalid input file: Header line '%s' not a .pkl header."%HeaderLine
        self.PrecursorMZ = float(Bits[0])
        self.PrecursorIntensity = float(Bits[1])
        self.Charge = int(Bits[2])
        # We hope to be called with an actual charge.  If we didn't get one at all, then guess 2.
        if not self.Charge:
            self.Charge = 2
        self.ParentMass = (self.PrecursorMZ * self.Charge) - (1.0078 * (self.Charge - 1))
        ##print "Prec %.2f times charge %s gives pm %s"%(self.PrecursorMZ, self.Charge, self.ParentMass)
        # All subsequent lines: Mass and intensity
        self.Peaks = []
        for FileLine in File.xreadlines():
            Bits = FileLine.split()
            if len(Bits) > 2:
                break
            self.ReadPeakDTALine(FileLine)
        File.close()
        self.Peaks.sort() # sort by mass
    def ReadPeaksFromFile(self, File, FileName):
        # Given a file 'blah123.dta', name the spectrum 'blah123'.
        self.Name = os.path.split(FileName)[1]
        (self.Name, FileExtension) = os.path.splitext(self.Name)
        FileExtension = FileExtension.lower()
        # Strip ".mzxml:444279" to ".mzxml":
        if FileExtension.find(":")!=-1:
            FileExtension = FileExtension[:FileExtension.find(":")]
        # Use the appropriate parser
        if FileExtension == ".pkl":
            self.ReadPeaksPKL(File)
        elif FileExtension == ".mzxml":
            self.ReadPeaksMZXML(File)
        elif FileExtension == ".mzdata":
            self.ReadPeaksMZData(File)
        elif FileExtension == ".mgf":
            self.ReadPeaksMGF(File)
        else:
            # default case: DTA
            self.ReadPeaksDTA(File)
    def ReadPeaks(self, FileName, FilePos = None):
        """
        Instantiator - Read a spectrum from a file.  Sets ParentMass, Charge,
        and Peaks list.  Doesn't filter, yet.
        """
        if FilePos == None:
            try:
                ColonBits = FileName.split(":")
                FilePos = int(ColonBits[-1])
                FileName = string.join(ColonBits[:-1], ":")
            except:
                FilePos = 0
        try:
            File = open(FileName, "rb")
        except:
            print "Error in ReadPeaks(): File '%s' couldn't be opened."%FileName
            traceback.print_exc()
            return
        File.seek(FilePos)
        self.ReadPeaksFromFile(File, FileName)
        self.FilePath = FileName
        self.FilePos = FilePos
        File.close()
    def ReadPeaksMZData(self, File):
        """
        Parse peaks from an .mzdata format file.  This format is slightly inferior
        to .mzxml, and not as commonly used.
        """
        ParseXML.GetSpectrumPeaksMZData(self, File)
        self.Charge = 2 # guess!
        self.ParentMass = (self.PrecursorMZ * self.Charge) - (self.Charge - 1) * 1.0078
        return 
        
    def ReadPeaksMZXML(self, File):
        """
        Parse peaks from an .mzXML format file.  Assumes we've already scanned to the
        desired file offset.
        """
        ParseXML.GetSpectrumPeaksMZXML(self, File)
        self.Charge = 2 # guess!
        self.ParentMass = (self.PrecursorMZ * self.Charge) - (self.Charge - 1) * 1.0078
        return 
    def DebugPrint(self, ShowPeaks = 0):
        """
        Print information on our spectrum, for debugging.
        """
        print "Spectrum '%s' has parent mass %f,\n   charge %f, and %d peaks"%(self.Name, self.ParentMass,
            self.Charge, len(self.Peaks))
        if self.CorrectPeptide:
            print " True parent peptide is: %s"%self.CorrectPeptide
        if ShowPeaks:
            for Peak in self.Peaks:
                if Peak.IonType:
                    print "  %f\t%f\t%s"%(Peak.Mass, Peak.Intensity, Peak.IonType.Name)
                else:
                    print "  %f\t%f"%(Peak.Mass, Peak.Intensity)
    def GetBestPeak(self, Mass, MaxIntensity = None, Epsilon = 1.0):
        "Used in labeling.  Find the best nearby peak whose intensity doesn't exceed our limit."
        if MaxIntensity == 0:
            return (None, None)
        BestPeak = None
        BestPeakError = None
        ClosestError = None
        for Peak in self.Peaks:
            Error = Peak.Mass - Mass
            if Error < -Epsilon:
                continue
            if Error > Epsilon:
                break
            if MaxIntensity and Peak.Intensity > MaxIntensity:
                continue # forbid neutral losses which are taller than the original            
            if (BestPeak == None or BestPeak.Intensity < Peak.Intensity):
                BestPeak = Peak
                BestPeakError = Error
        return (BestPeak, BestPeakError)
        
    def GetPeak(self, Mass, Epsilon = 1.0):
        """
        Get the closest peak to the specified mass, with a maximum error of Epsilon.
        """
        ClosestPeak = None
        ClosestError = None
        for Peak in self.Peaks:
            Error = abs(Peak.Mass - Mass)
            if Error < Epsilon:
                if (ClosestPeak == None or ClosestError > Error):
                    ClosestPeak = Peak
                    ClosestError = Error
            if Peak.Mass > Mass:
                break
        return ClosestPeak
    def GetPRMPeak(self, Mass, Epsilon = 1.0):
        """
        Get the closest peak to the specified mass, with a maximum error of Epsilon.
        """
        ClosestPeak = None
        ClosestError = None
        for Peak in self.PRMPeaks:
            Error = abs(Peak.Mass - Mass)
            if Error < Epsilon:
                if (ClosestPeak == None or ClosestError > Error):
                    ClosestPeak = Peak
                    ClosestError = Error
            if Peak.Mass > Mass:
                break
        return ClosestPeak
    def GetPRMPeaks(self, Mass, Epsilon = 1.0):
        """
        Get all peaks within Epsilon of Mass
        """
        Peaks = []
        for Peak in self.PRMPeaks:
            Error = abs(Peak.Mass - Mass)
            if Error < Epsilon:
                Peaks.append(Peak)
            if Peak.Mass > Mass:
                break
        return Peaks
    def AssignIonTypesFromPeptide(self):
        """
        Assign ion types to our peaks, based on the CorrectPeptide.
        """
        # The true PRMPeaks are sums of peptide masses.  Iterate
        # over the length of the peptide:
        LeftMass = 0
        for Index in range(0, len(self.CorrectPeptide.Aminos)):
            LeftMass += Global.AminoMass[self.CorrectPeptide.Aminos[Index]]
            LeftMass += Global.FixedMods.get(self.CorrectPeptide.Aminos[Index], 0)
            # For this PRMPeak, look for all the possible spectral peaks
            # corresponding to the various ion types:
            for IonType in AllIons:
                Mass = IonType.GetPeakMass(LeftMass, self.ParentMass)
                Peak = self.GetPeak(Mass, 1.0)
                if Peak:
                    Peak.IonType = IonType
                    Peak.Pep = self.CorrectPeptide[:Index+1]
    def ApplyWindowFilter(self, RegionCutoffs, WindowSizes, MaxRankInclusive):
        """
        Apply this window-fiter to our peaks.  RegionCutoffs describe the edges
        of "early", "medium" and "late" spectral portions; WindowSizes are the
        sizes (in AMUs) of windows for these portions.  MaxRankInclusive is the
        worst rank to keep.  
        """
        #print "Apply window:", WindowSizes, RegionCutoffs, MaxRankInclusive
        GoodPeaks = []
        # List of region-edges:
        Borders = []
        for Cutoff in RegionCutoffs:
            Borders.append(self.ParentMass * Cutoff)
        NextBorderIndex = 0
        LastBorderIndex = len(RegionCutoffs)
        WindowIndex = 0
        BadPeakIntensityList = []
        for Peak in self.Peaks:
            while (NextBorderIndex < LastBorderIndex and Peak.Mass > Borders[NextBorderIndex]):
                NextBorderIndex += 1
                WindowIndex += 1
            WindowSize = WindowSizes[WindowIndex]
            MinMass = Peak.Mass - WindowSize/2
            MaxMass = Peak.Mass + WindowSize/2
            List = []
            for OtherPeak in self.Peaks:
                if OtherPeak.Mass > MaxMass:
                    break
                if OtherPeak.Mass > MinMass:
                    List.append((OtherPeak.Intensity, OtherPeak))
            List.sort()
            List.reverse() # best to worst
            if (len(List) < MaxRankInclusive+1) or (Peak.Intensity >= List[MaxRankInclusive][0]):
                GoodPeaks.append(Peak)
            else:
                BadPeakIntensityList.append(Peak.Intensity)
        #print "Kept %d of %d original peaks."%(len(GoodPeaks), len(self.Peaks))
        self.Peaks = GoodPeaks
        if len(BadPeakIntensityList):
            BadPeakIntensityList.sort() 
            return BadPeakIntensityList[len(BadPeakIntensityList)/2]
        else:
            return -1
    def FilterPeaks(self, WindowSize = 50, PeakCount = 6):
        self.ApplyWindowFilter([], (WindowSize,), PeakCount - 1)       
    def WritePeaks(self, FilePath):
        "Write out a .dta file."
        File = open(FilePath, "w")
        File.write("%f\t%d\n"%(self.ParentMass, self.Charge))
        for Peak in self.Peaks:
            File.write("%f\t%f\n"%(Peak.Mass, Peak.Intensity))
        File.close()
    def WritePKLPeaks(self,FilePath):
        """"
        Append to the end of a .pkl file.  Note this APPENDS a file.
        if no precursor intensity is known, then we say zero.  I hope that does not break things.
        """
        FileHandle = open(FilePath, "a")
        if self.PrecursorIntensity:
            FileHandle.write("%s %s %s\n"%(self.PrecursorMZ,self.PrecursorIntensity,self.Charge))
        else:
            FileHandle.write("%s 0.0 %s\n"%(self.PrecursorMZ,self.Charge))
        for Peak in self.Peaks:
            FileHandle.write("%f\t%f\n"%(Peak.Mass, Peak.Intensity))
        FileHandle.write("\n") #need a blank line to separate different scans
        FileHandle.close
    def WriteMGFPeaks(self, TheFile, Title = "Spectrum", ScanNumber = None):
        """
        Append to the end of an mgf file.  Pass in an open file, or
        (as a string) the path of a file to be APPENDED to.
        """
        if type(TheFile) == type(""):
            File = open(FilePath, "a")
        else:
            File = TheFile
        File.write("BEGIN IONS\n")
        File.write("TITLE=%s\n"%Title)
        if ScanNumber != None:
            File.write("SCAN=%s\n"%ScanNumber)
        File.write("CHARGE=%d\n"%self.Charge)
        File.write("PEPMASS=%f\n"%self.PrecursorMZ)
        for Peak in self.Peaks:
            File.write("%f\t%f\n"%(Peak.Mass, Peak.Intensity))
        File.write("END IONS\n")
        if type(TheFile) == type(""):
            File.close()
    def WriteMZXMLPeaks(self, File, ScanNumber):
        PeakCount = len(self.Peaks)
        Str = """<scan num="%s" msLevel="2" peaksCount="%s" polarity="+" scanType="Full" lowMz="125" highMz="2000" """%(ScanNumber, PeakCount)
        Str += """\n<precursorMz """
        if self.PrecursorIntensity:
            Str += """ precursorIntensity = "%.2f" """%self.PrecursorIntensity
        Str += ">%.5f</precursorMz>\n"%self.PrecursorMZ
        Str += """\n<peaks precision="32" byteOrder="network" pairOrder="m/z-int">"""
        PeakString = ""
        for Peak in self.Peaks:
            PeakString += struct.pack(">ff", Peak.Mass, Peak.Intensity)
        PeakString = B64Encode(PeakString)
        Str += PeakString
        Str += "</peaks>\n</scan>\n"
        File.write(Str + "\n")
        
    def GetTopUnexplainedPeak(self):
        TopUXRank = len(self.Peaks)
        for Peak in self.Peaks:
            if Peak.IonType == None and Peak.IntensityRank < TopUXRank:
                TopUXRank = Peak.IntensityRank
        return TopUXRank
    def FindIsotopicPeaks(self):
        for PeakIndex in range(len(self.Peaks)):
            Peak = self.Peaks[PeakIndex]
            RoundMass = int(round(Peak.Mass))
            ExpectedFraction = Global.IsotopeWeights.get(RoundMass, None)
            if ExpectedFraction==None:
                continue
            for IsotopePeakIndex in range(PeakIndex+1, len(self.Peaks)):
                OtherPeak = self.Peaks[IsotopePeakIndex]
                if OtherPeak.Mass < Peak.Mass + 0.8:
                    continue
                if OtherPeak.Mass > Peak.Mass + 1.2:
                    break
                Fraction = OtherPeak.Intensity / Peak.Intensity
                # magic numbers ahoy:
                if abs(Fraction - ExpectedFraction) < 0.5 or (abs(Fraction - ExpectedFraction) < 0.8 and OtherPeak.Mass > Peak.Mass + 0.9 and OtherPeak.Mass < Peak.Mass + 1.1): 
                    OtherPeak.IsPlausibleIsotopicPeak = 1
                    Peak.HasPlausibleIsotopicPeak = 1
    def GetExplainedIntensity(self):
        """
        Callable *after* the spectrum has been labeled.  Returns the percentage
        of total spectral intensity that has been explained by labels.  All
        things being equal, a candidate peptide with a higher explained
        intensity is BETTER.
        """
        TotalIntensity = 0
        ExplainedIntensity = 0
        for Peak in self.Peaks:
            TotalIntensity += Peak.Intensity
            if Peak.IonType != None:
                ExplainedIntensity += Peak.Intensity
            #print "%s\t%s\t%s\t%s\t%s\t"%(Peak.Mass, Peak.Intensity, Peak.IonType, Peak.PeptideIndex, Peak.RescueFlag)
        return ExplainedIntensity / float(max(1, TotalIntensity))
    def GetExplainedIons(self, Peptide, DynamicRangeMin = 150, DynamicRangeMax = 2000):
        "Return the percentage of b and y peaks present."
        Annotated = {}
        
        PhosphorylationFlag = 0
        PhosB = [0]*40
        PhosY = [0]*40
        for (Pos, ModList) in Peptide.Modifications.items():
            for Mod in ModList:
                if Mod.Name == "Phosphorylation":
                    PhosphorylationFlag = 1
        for Peak in self.Peaks:
            if Peak.IonType:
                Annotated[(Peak.IonType.Name, Peak.PeptideIndex)] = 1
        Count = 0
        Present = 0
        TotalCutPresent = 0
        PM = 19 + Peptide.Masses[-1]
        for Index in range(len(Peptide.Masses)):
            CutPresent = 0
            BMass = Peptide.Masses[Index] + 1.0078
            if BMass > DynamicRangeMin and BMass < DynamicRangeMax:
                Count += 1
                BPresent = Annotated.get(("b", Index),0)
                BPresent |= Annotated.get(("b-p", Index),0)
                Present += BPresent
                CutPresent |= BPresent
            YMass = PM - Peptide.Masses[Index]
            if YMass > DynamicRangeMin and YMass < DynamicRangeMax:
                Count += 1
                YPresent = Annotated.get(("y", len(Peptide.Aminos) - Index), 0)
                YPresent |= Annotated.get(("y-p", Index),0)
                Present += YPresent
                CutPresent |= YPresent
            # Count the CUT POINTS that are witnessed:
            if (Index and Index<len(Peptide.Masses)-1) and CutPresent:
                TotalCutPresent += 1
        CutCount = len(Peptide.Masses) - 1
        #print Peptide.Aminos, "%s cut points of %s"%(TotalCutPresent, CutCount)
        return (Present, Count, Present / max(1, float(Count)),
                TotalCutPresent, CutCount, TotalCutPresent/max(1, float(CutCount)))
            
        
    def GetExplainedPeaks(self, MaxRank = 24):
        """
        Returns the percentage of the top n peaks that have been explained
        by peak labeling, where n = MaxRank.  The output of GetExplainedPeaks()
        should be high for a good candidate peptide.
        """
        TotalGoodPeaks = 0
        ExplainedGoodPeaks = 0
        for Peak in self.Peaks:
            if Peak.IntensityRank <= MaxRank:
                TotalGoodPeaks += 1
                #print Peak.IntensityRank, Peak.Mass, Peak.IonType
                if Peak.IonType != None:
                    ExplainedGoodPeaks += 1
        return ExplainedGoodPeaks / float(max(1, TotalGoodPeaks))
    def GetLogMeanStdev(self):
        """computes the mean and standard deviation of the peak intensities
        This can be done at any time before or after filtering.  just
        make sure that you know what it means.
        This computes things based on the LOG intensity values
        """
        IntensitySum = 0.0
        NumPeaks = len(self.Peaks)
        for Peak in self.Peaks:
            IntensitySum += math.log(Peak.Intensity)
        Mean = IntensitySum / NumPeaks
        VarSum = 0.0
        for Peak in self.Peaks:
            Diff = math.log(Peak.Intensity) - Mean
            VarSum += Diff*Diff
        Variance = VarSum / NumPeaks
        Stdev = math.sqrt(Variance)
        return (Mean,Stdev)
    



#Title:          SelectProteins.py
#Author:         Stephen Tanner, Samuel Payne, Natalie Castellana, Pavel Pevzner, Vineet Bafna
#Created:        2005
# Copyright 2007,2008,2009 The Regents of the University of California
# All Rights Reserved
#
# Permission to use, copy, modify and distribute any part of this
# program for educational, research and non-profit purposes, by non-profit
# institutions only, without fee, and without a written agreement is hereby
# granted, provided that the above copyright notice, this paragraph and
# the following three paragraphs appear in all copies.
#
# Those desiring to incorporate this work into commercial
# products or use for commercial purposes should contact the Technology
# Transfer & Intellectual Property Services, University of California,
# San Diego, 9500 Gilman Drive, Mail Code 0910, La Jolla, CA 92093-0910,
# Ph: (858) 534-5815, FAX: (858) 534-7345, E-MAIL:invent@ucsd.edu.
#
# IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
# INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE, EVEN
# IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.
#
# THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
# OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.  THE UNIVERSITY OF CALIFORNIA MAKES NO
# REPRESENTATIONS AND EXTENDS NO WARRANTIES OF ANY KIND, EITHER IMPLIED OR
# EXPRESS, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, TRADEMARK OR OTHER RIGHTS.


"""
Helper class for FDRUtils.py and FDR.py:
Given an f-score cutoff, select a parsimonious collection of proteins which
account for most / all of the annotations.
"""
import os
import sys
import traceback
import struct
import ResultsParser
from Utils import *
Initialize()

class ProteinSelector(ResultsParser.ResultsParser):
    def __init__(self):
        self.PeptideDict = {} # aminos -> location list
        self.ProteinPeptideCounts = {}
        self.ProteinSpectrumCounts = {}
        #self.SharedProteinPeptides = {}
        #self.SharedProteinSpectra = {}
        self.PeptideSpectrumCounts = {}
        self.ProteinPeptides = {} # Protein -> List of aminos
        self.ProteinNames = {}
        self.ProteinSequences = {}
        self.MQScoreWeight = 0.3
        self.DeltaScoreWeight = 1.5
        self.MinimumPeptideLength = 7
        self.BestScoresByPeptide = {}
        self.PValueCutoff = None
        self.MaxFileLines = None
        # if RetainRepresentativeCount is set, then we remember the
        # best n spectra for a particular annotation in the dictionary
        # self.BestRepresentatives
        self.RetainRepresentativeCount = None
        self.BestRepresentatives = {}
        self.AnnotationSpectrumCounts = {}
        self.FScoreCutoff2 = None
        self.FScoreCutoff3 = None
        self.Columns = ResultsParser.Columns()
        ResultsParser.ResultsParser.__init__(self)

    def FindPeptideLocations(self, Aminos):
        PrevPos = -1
        
        LocationList = []
        while (1):
            Pos = self.DB.find(Aminos, PrevPos + 1)
            if Pos == -1:
                break
            
            # Which protein does Pos lie in?
            LowIndex = 0
            HighIndex = len(self.ProteinPos) - 1
            # Pos >= ProteinPos[LowIndex] and Pos < ProteinPos[HighIndex]
            # Special case - last protein:
            
            if Pos >= self.ProteinPos[HighIndex]:
                ProteinID = HighIndex
                ResidueNumber = Pos - self.ProteinPos[HighIndex]
               
            else:
                while (1):
                    if LowIndex+1==HighIndex:
                        ProteinID = LowIndex
                        ResidueNumber = Pos - self.ProteinPos[LowIndex]
                        break
                    MidIndex = (LowIndex + HighIndex) / 2
                    if Pos >= self.ProteinPos[MidIndex]:
                        LowIndex = MidIndex
                    else:
                        HighIndex = MidIndex
            LocationList.append((ProteinID, ResidueNumber))
            PrevPos = Pos
        return LocationList
    def OldFindPeptideLocations(self, Aminos):
        LocationList = []
        #print "Find locations for %s..."%Aminos
        for (ID, Sequence) in self.ProteinSequences.items():
            Pos = Sequence.find(Aminos)
            if Pos != -1:
                LocationList.append((ID, Pos))
                #print "Found at pos %s in %s"%(Pos, ID)
        if len(LocationList) == 0:
            print "*** WARNING: Peptide '%s' not found in the database."%Aminos
        return LocationList
    def LoadDB(self, DBPath):
        DBFile = open(DBPath, "rb")
        self.DB = DBFile.read()
        DBFile.close()
        IndexPath = os.path.splitext(DBPath)[0] + ".index"
        IndexFile = open(IndexPath, "rb")
        BlockSize = struct.calcsize("<qi80s")
        ID = 0
        PrevID = None
        self.ProteinPos = []
        while (1):
            Block = IndexFile.read(BlockSize)
            if not Block:
                break
            Info = struct.unpack("<qi80s", Block)
            Name = Info[2]
            NullPos = Name.find("\0")
            if NullPos !=- 1:
                Name = Name[:NullPos]
            self.ProteinNames[ID]= Name
            StartPos = Info[1]
            self.ProteinPos.append(StartPos)
            if PrevID != None:
                self.ProteinSequences[PrevID] = self.DB[self.ProteinPos[PrevID]:StartPos - 1]
            PrevID = ID
            ID += 1
        self.ProteinSequences[PrevID] = self.DB[self.ProteinPos[PrevID]:]
    def LoadMultipleDB(self, DBPathList):
        """" Given a list of DB pathnames, load all the corresponding DB """
        ID = 0
        self.DB = ""
        self.ProteinPos = []
        for DBPath in DBPathList:
            print "loading %s"%DBPath
            DBFile = open(DBPath, "rb")
            OldDB = self.DB
            self.DB += DBFile.read()    # concatenate all DBs sequentially
            DBFile.close()
            IndexPath = os.path.splitext(DBPath)[0] + ".index"
            IndexFile = open(IndexPath, "rb")
            BlockSize = struct.calcsize("<qi80s")
            PrevID = None
            while (1):
                Block = IndexFile.read(BlockSize)
                if not Block:
                    break
                Info = struct.unpack("<qi80s", Block)
                Name = Info[2]
                NullPos = Name.find("\0")
                if NullPos !=- 1:
                    Name = Name[:NullPos]
                self.ProteinNames[ID]= Name
                StartPos = Info[1] + len(OldDB) # adjust StartPos for adding a new DB                
                self.ProteinPos.append(StartPos)
                if PrevID != None:
                    self.ProteinSequences[PrevID] = self.DB[self.ProteinPos[PrevID]:StartPos - 1]
                    
                PrevID = ID
                ID += 1
            self.ProteinSequences[PrevID] = self.DB[self.ProteinPos[PrevID]:]
            
    def OldLoadDB(self, DBPath):
        """
        Load the database, popluating self.ProteinSequences
        """
        print "LoadDB(%s)"%DBPath
        IndexPath = os.path.splitext(DBPath)[0] + ".index"
        IndexFile = open(IndexPath, "rb")
        DBFile = open(DBPath, "rb")
        BlockSize = struct.calcsize("<qi80s")
        PrevName = None
        PrevID = None
        PrevStartPos = None
        ID = 0
        while (1):
            Block = IndexFile.read(BlockSize)
            if not Block:
                break
            Info = struct.unpack("<qi80s", Block)
            Name = Info[2]
            NullPos = Name.find("\0")
            if NullPos !=- 1:
                Name = Name[:NullPos]
            StartPos = Info[1]
            self.ProteinNames[ID] = Name
            if PrevName != None:
                DBFile.seek(PrevStartPos)
                Sequence = DBFile.read(StartPos - PrevStartPos)
                Sequence = Sequence.replace("*", "")
                self.ProteinSequences[PrevID] = Sequence
            PrevName = Name
            PrevID = ID
            PrevStartPos = StartPos
            ID += 1
        if PrevName != None:
            DBFile.seek(StartPos)
            Sequence = DBFile.read()
            self.ProteinSequences[PrevID] = Sequence
            #self.ProteinNames[PrevID] = Name

        DBFile.close()
        IndexFile.close()
    def ChooseProteins(self):
        """
        Iteratively select proteins which account for all the peptides.
        """
        self.SelectedProteins = {} # Protein -> (Peptides, Spectra)
        self.PeptideProteins = {} # Peptide -> final selection of protein
        print "\n\n\n"
        print "CHOOSE PROTEINS:"
        for (Peptide, SpectrumCount) in self.PeptideSpectrumCounts.items():
            for (ProteinID, Pos) in self.PeptideDict[Peptide]:
                self.ProteinSpectrumCounts[ProteinID] = self.ProteinSpectrumCounts.get(ProteinID, 0) + SpectrumCount

        print "Loaded %s peptides and %s proteins"%(len(self.PeptideSpectrumCounts.keys()), len(self.ProteinSpectrumCounts.keys()))
        while (1):
            BestCandidate = None
            BestScore = None
            for Protein in self.ProteinPeptideCounts.keys():
                if self.SelectedProteins.has_key(Protein):
                    continue
                PeptideCount = self.ProteinPeptideCounts[Protein]
                SpectrumCount = self.ProteinSpectrumCounts.get(Protein, 0)
                Score = (PeptideCount, SpectrumCount)
                #print Protein, Score
                if Score > BestScore or (Score == BestScore and self.ProteinNames[Protein] < self.ProteinNames[BestCandidate]):
                    BestScore = Score
                    BestCandidate = Protein
                    #print "New Best %s, score %s"%(BestCandidate,BestScore)
            if not BestScore:
                break
            (PeptideCount, SpectrumCount) = BestScore
            if PeptideCount == 0:
                break
            #%%%
            print "Accept protein %s (%s)\n  Gets %s peptides, %s spectra"%(BestCandidate, self.ProteinNames[BestCandidate], PeptideCount, SpectrumCount)
            
            self.SelectedProteins[BestCandidate] = BestScore
            # Lay claim to all the (not-yet-claimed) peptides:
            for Peptide in self.ProteinPeptides[BestCandidate]:
                #print Peptide
                if not self.PeptideProteins.has_key(Peptide):
                    
                    self.PeptideProteins[Peptide] = BestCandidate
                    # Other proteins (if not already accepted) lose a peptide, and some spectra:
                    for (OtherProtein, Pos) in self.PeptideDict[Peptide]:
                        if self.SelectedProteins.has_key(OtherProtein):
                            continue
                        
                        self.ProteinPeptideCounts[OtherProtein] -= 1
                        self.ProteinSpectrumCounts[OtherProtein] = self.ProteinSpectrumCounts.get(OtherProtein, 0) - self.PeptideSpectrumCounts[Peptide]
        # Sanity check - the selected proteins have peptides, the unselected proteins have 0
        for Protein in self.ProteinPeptideCounts.keys():
            ProteinName = self.ProteinNames[Protein]
            PeptideCount = self.ProteinPeptideCounts[Protein]
            SpectrumCount = self.ProteinSpectrumCounts.get(Protein, 0)
            if self.SelectedProteins.has_key(Protein) and PeptideCount <= 0:
                print "** Warning: Selected protein %s (%s) has %s peptides!"%(Protein, ProteinName, PeptideCount)
            if not self.SelectedProteins.has_key(Protein) and PeptideCount != 0:
                print "** Warning: Unelected protein %s (%s) has %s peptides!"%(Protein, ProteinName, PeptideCount)
    def ParseAnnotations(self, FileName):
        """
        Parse annotations, remembering all protein locations for each peptide.
        """
        print "Parse %s..."%FileName
        File = open(FileName, "rb")
        OldSpectrum = None
        Stub = os.path.split(FileName)[1]
        LineNumber = 0
        for FileLine in File:
            LineNumber += 1
            if LineNumber % 100 == 0:
                print "%s %s..."%(Stub, LineNumber)
                if self.MaxFileLines != None and LineNumber >= self.MaxFileLines:
                    return # Quick-parse, for debugging only!
            
            
            if FileLine[0] == "#":
                self.Columns.initializeHeaders(FileLine)
                continue
            if not FileLine.strip():
                continue                        
            Bits = FileLine.split("\t")
            try:
                Spectrum = (Bits[self.Columns.getIndex("SpectrumFile")], Bits[self.Columns.getIndex("Scan#")])
            except:
                continue # header line
            if Spectrum == OldSpectrum:
                continue
            OldSpectrum = Spectrum
            try:
                MQScore = float(Bits[self.Columns.getIndex("MQScore")])
                DeltaScore = float(Bits[self.Columns.getIndex("DeltaScoreOther")])
                Charge = int(Bits[self.Columns.getIndex("Charge")])
            except:
                traceback.print_exc()
                print Bits
                continue
            # Apply a threshold: EITHER f-score cutoff (default) OR p-value cutoff
            
            if self.PValueCutoff != None:
                
                try:
                    PValue = float(Bits[self.Columns.getIndex("InspectFDR")])
                except:
                    traceback.print_exc()
                    print Bits
                    continue
                PeptideScore = (-PValue, MQScore)
                if PValue > self.PValueCutoff:
                    
                    continue
            else:
                if Charge < 3:
                    WeightedScore = self.MQScoreWeight * MQScore + self.DeltaScoreWeight * (DeltaScore / self.MeanDeltaScore2)
                    if WeightedScore < self.FScoreCutoff2:
                        
                        continue
                else:
                    WeightedScore = self.MQScoreWeight * MQScore + self.DeltaScoreWeight * (DeltaScore / self.MeanDeltaScore3)
                    if WeightedScore < self.FScoreCutoff3:
                        
                        continue
                PeptideScore = WeightedScore
                
            
            try:
                Peptide = GetPeptideFromModdedName(Bits[self.Columns.getIndex("Annotation")])
            except:
                
                continue
            if len(Peptide.Aminos) < self.MinimumPeptideLength:
                
                continue
            # Remember this peptide:
            
            if not self.PeptideDict.get(Peptide.Aminos):
                
                # It's a new peptide!  Figure out where it falls in the database:
                LocationList = self.FindPeptideLocations(Peptide.Aminos)
                for (Protein, Pos) in LocationList:
                    if not self.ProteinPeptides.has_key(Protein):
                        self.ProteinPeptides[Protein] = []
                    #print "Adding peptide %s to protein %s '%s':"%(Peptide.Aminos,Protein,self.ProteinNames[Protein])
                    self.ProteinPeptides[Protein].append(Peptide.Aminos)
               
                self.PeptideDict[Peptide.Aminos] = LocationList
                for (ProteinNumber, Dummy) in LocationList:
                    self.ProteinPeptideCounts[ProteinNumber] = self.ProteinPeptideCounts.get(ProteinNumber, 0) + 1
            else:
                # We've seen this peptide before:
                LocationList = self.PeptideDict[Peptide.Aminos]
            OldScore = self.BestScoresByPeptide.get(Peptide.Aminos, -9999)
            self.BestScoresByPeptide[Peptide.Aminos] = max(PeptideScore, OldScore)
            self.PeptideSpectrumCounts[Peptide.Aminos] = self.PeptideSpectrumCounts.get(Peptide.Aminos, 0) + 1
            ##############################################################
            # Populate self.BestRepresentative, if requested:
            if self.RetainRepresentativeCount:
                Peptide.MQScore = MQScore
                Peptide.PValue = PValue
                Peptide.SpectrumFilePath = Bits[0]
                Peptide.ScanNumber = int(Bits[1])
                Peptide.SpectrumFilePos = int(Bits[self.Columns.getIndex("SpecFilePos")])
                Key = Peptide.GetFullModdedName()
                RepresentativeList = self.BestRepresentatives.get(Key, [])
                Tuple = (PeptideScore, Peptide)
                RepresentativeList.append(Tuple)
                RepresentativeList.sort()
                self.BestRepresentatives[Key] = RepresentativeList[-self.RetainRepresentativeCount:]
                self.AnnotationSpectrumCounts[Key] = self.AnnotationSpectrumCounts.get(Key, 0) + 1
                

if __name__ == "__main__":
    # Test
    Bob = ProteinSelector()
    Bob.LoadDB("database\DictyCommon.Aug28.FS2.trie")
    print Bob.FindPeptideLocations("GTVESEMAEQDSLLNKLNK")
    print Bob.FindPeptideLocations("TSEGDFTLLLGQIVDNQIGDLNKSG")
    print Bob.FindPeptideLocations("YAVFAPGLADVVIEVVAK")
